import os
import platform
import shutil
from logging import getLogger
from pathlib import Path
from typing import Optional, List, Dict
from dataclasses import asdict

import vdf

from rare.utils.paths import image_icon_path, image_wide_path, image_tall_path, desktop_icon_path, get_rare_executable
from rare.models.steam import SteamUser, SteamShortcut

if platform.system() == "Windows":
    # noinspection PyUnresolvedReferences
    import winreg  # pylint: disable=E0401

logger = getLogger("SteamShortcuts")

steam_client_install_paths = [os.path.expanduser("~/.local/share/Steam")]


def find_steam() -> Optional[str]:
    if platform.system() == "Windows":
        # Find the Steam install directory or raise an error
        try:  # 32-bit
            key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Valve\\Steams")
        except FileNotFoundError:
            try:  # 64-bit
                key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Wow6432Node\\Valve\\Steam")
            except FileNotFoundError as e:
                return None
        return winreg.QueryValueEx(key, "InstallPath")[0]
    # return the first valid path
    elif platform.system() in {"Linux", "FreeBSD"}:
        for path in steam_client_install_paths:
            if os.path.isdir(path) and os.path.isfile(os.path.join(path, "steam.sh")):
                return path
    return None


def find_steam_users(steam_path: str) -> List[SteamUser]:
    _users = []
    vdf_path = os.path.join(steam_path, "config", "loginusers.vdf")
    if not os.path.exists(vdf_path):
        return _users
    with open(vdf_path, "r", encoding="utf-8") as f:
        users = vdf.load(f).get("users", {})
    for long_id, user in users.items():
        _users.append(SteamUser(long_id, user))
    return _users


def _load_shortcuts(steam_path: str, user: SteamUser) -> Dict[str, SteamShortcut]:
    _shortcuts = {}
    vdf_path = os.path.join(steam_path, "userdata", str(user.short_id), "config", "shortcuts.vdf")
    if not os.path.exists(vdf_path):
        return _shortcuts
    with open(vdf_path, "rb") as f:
        shortcuts = vdf.binary_load(f).get("shortcuts", {})
    for idx, shortcut in shortcuts.items():
        _shortcuts[idx] = SteamShortcut.from_dict(shortcut)
    return _shortcuts


def _save_shortcuts(steam_path: str, user: SteamUser, shortcuts: Dict[str, SteamShortcut]) -> None:
    _shortcuts = {k: asdict(v) for k, v in shortcuts.items()}
    vdf_path = os.path.join(steam_path, "userdata", str(user.short_id), "config", "shortcuts.vdf")
    with open(vdf_path, "wb") as f:
        vdf.binary_dump({"shortcuts": _shortcuts}, f)


__steam_dir: Optional[str] = None
__steam_user: Optional[SteamUser] = None
__steam_shortcuts: Optional[Dict] = None


def steam_shortcuts_supported() -> bool:
    return __steam_dir is not None and __steam_user is not None and __steam_shortcuts is not None


def load_steam_shortcuts():
    global __steam_shortcuts, __steam_dir, __steam_user

    if __steam_shortcuts is not None:
        return

    steam_dir = find_steam()
    if not steam_dir:
        logger.error("Failed to find Steam install directory")
        return
    __steam_dir = steam_dir

    steam_users = find_steam_users(steam_dir)
    if not steam_users:
        logger.error("Failed to find any Steam users")
        return
    else:
        steam_user = next(
            filter(lambda x: x.most_recent, steam_users),
            sorted(steam_users, key=lambda x: x.last_login, reverse=True)[0]
        )
        logger.info(
            "Found most recently logged-in user %s(%s) (%s)",
            steam_user.account_name, steam_user.persona_name, steam_user.last_login
        )
    __steam_user = steam_user

    __steam_shortcuts = _load_shortcuts(steam_dir, steam_user)


def save_steam_shortcuts():
    logger.info(
        "%s Steam shortcuts for user %s(%s)",
        "Saving" if __steam_shortcuts else "Removing",
        __steam_user.account_name,
        __steam_user.persona_name
    )
    _save_shortcuts(__steam_dir, __steam_user, __steam_shortcuts)


def steam_shortcut_exists(app_name: str) -> bool:
    return SteamShortcut.calculate_appid(app_name) in {s.appid for s in __steam_shortcuts.values()}


def remove_steam_shortcut(app_name: str) -> Optional[SteamShortcut]:
    global __steam_shortcuts

    if not steam_shortcut_exists(app_name):
        logger.error("Game %s doesn't have an associated Steam shortcut", app_name)
        return None

    appid = SteamShortcut.calculate_appid(app_name)
    removed = next(filter(lambda item: item[1].appid == appid, __steam_shortcuts.items()))
    shortcuts = dict(filter(lambda item: item[1].appid != appid, __steam_shortcuts.items()))
    __steam_shortcuts = shortcuts
    return removed[1]


def add_steam_shortcut(app_name: str, app_title: str) -> SteamShortcut:
    global __steam_shortcuts

    if steam_shortcut_exists(app_name):
        logger.info("Removing old Steam shortcut for %s", app_name)
        remove_steam_shortcut(app_name)

    command = get_rare_executable()
    arguments = ["launch", app_name]
    if len(command) > 1:
        arguments = command[1:] + arguments
    shortcut = SteamShortcut.create(
        app_name=app_name,
        app_title=f"{app_title} (Rare)",
        executable=command[0],
        start_dir=os.path.dirname(command[0]),
        icon=desktop_icon_path(app_name).as_posix(),
        launch_options=arguments,
    )

    key = int(max(__steam_shortcuts.keys(), default="0"))
    __steam_shortcuts[str(key+1)] = shortcut
    return shortcut


def add_steam_coverart(app_name: str, shortcut: SteamShortcut):
    steam_grid_dir = os.path.join(__steam_dir, "userdata", str(__steam_user.short_id), "config", "grid")
    if not os.path.exists(steam_grid_dir):
        os.mkdir(steam_grid_dir)
    shutil.copy(image_wide_path(app_name), os.path.join(steam_grid_dir, shortcut.game_hero))
    shutil.copy(image_icon_path(app_name), os.path.join(steam_grid_dir, shortcut.game_logo))
    shutil.copy(image_wide_path(app_name), os.path.join(steam_grid_dir, shortcut.grid_wide))
    shutil.copy(image_tall_path(app_name), os.path.join(steam_grid_dir, shortcut.grid_tall))


def remove_steam_coverart(shortcut: SteamShortcut):
    steam_grid_dir = os.path.join(__steam_dir, "userdata", str(__steam_user.short_id), "config", "grid")
    if not os.path.exists(steam_grid_dir):
        logger.warning("Path does not exist %s", steam_grid_dir)
        return
    Path(steam_grid_dir).joinpath(shortcut.game_hero).unlink(missing_ok=True)
    Path(steam_grid_dir).joinpath(shortcut.game_logo).unlink(missing_ok=True)
    Path(steam_grid_dir).joinpath(shortcut.grid_wide).unlink(missing_ok=True)
    Path(steam_grid_dir).joinpath(shortcut.grid_tall).unlink(missing_ok=True)


if __name__ == "__main__":

    from pprint import pprint

    load_steam_shortcuts()

    print(__steam_dir)
    print(__steam_user)
    print(__steam_shortcuts)

    def print_shortcuts():
        for k, s in __steam_shortcuts.items():
            print({k: asdict(s)})
            print(vars(s))
        print()

    print_shortcuts()

    add_steam_shortcut("test1", "Test1")
    add_steam_shortcut("test2", "Test2")
    add_steam_shortcut("test3", "Test3")
    add_steam_shortcut("test1", "Test1")

    remove_steam_shortcut("test2")

    print_shortcuts()



import platform as pf
import shlex
import shutil
from logging import getLogger
from typing import Optional, Tuple, Iterable

from PySide6.QtCore import Signal, QSize, Qt, QMimeData, Slot, QObject, QEvent
from PySide6.QtGui import (
    QDrag,
    QDropEvent,
    QDragEnterEvent,
    QDragMoveEvent,
    QFont,
    QAction,
    QMouseEvent,
    QShowEvent,
    QResizeEvent,
)
from PySide6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QFrame,
    QMessageBox,
    QSizePolicy,
    QWidget,
    QScrollArea,
    QMenu,
    QPushButton,
    QLineEdit,
    QVBoxLayout,
    QComboBox,
)

from rare.models.wrapper import Wrapper
from rare.shared import RareCore
from rare.utils.misc import qta_icon
from rare.widgets.dialogs import ButtonDialog, game_title

if pf.system() in {"Linux", "FreeBSD"}:
    from rare.utils.compat import steam

logger = getLogger("WrapperSettings")


class WrapperEditDialog(ButtonDialog):
    result_ready = Signal(bool, str)

    def __init__(self, parent=None):
        super(WrapperEditDialog, self).__init__(parent=parent)

        self.line_edit = QLineEdit(self)
        self.line_edit.textChanged.connect(self.__on_text_changed)

        self.widget_layout = QVBoxLayout()
        self.widget_layout.addWidget(self.line_edit)

        self.setCentralLayout(self.widget_layout)

        self.accept_button.setText(self.tr("Save"))
        self.accept_button.setIcon(qta_icon("fa.edit"))
        self.accept_button.setEnabled(False)

        self.result: Tuple = ()

    def setup(self, wrapper: Wrapper):
        header = self.tr("Edit wrapper")
        self.setWindowTitle(header)
        self.setSubtitle(game_title(header, wrapper.name))
        self.line_edit.setText(wrapper.as_str)

    @Slot(str)
    def __on_text_changed(self, text: str):
        self.accept_button.setEnabled(bool(text))

    def done_handler(self):
        self.result_ready.emit(*self.result)

    def accept_handler(self):
        self.result = (True, self.line_edit.text())

    def reject_handler(self):
        self.result = (False, self.line_edit.text())


class WrapperAddDialog(WrapperEditDialog):
    def __init__(self, parent=None):
        super(WrapperAddDialog, self).__init__(parent=parent)
        self.combo_box = QComboBox(self)
        self.combo_box.addItem("None", "")
        self.combo_box.currentIndexChanged.connect(self.__on_index_changed)
        self.widget_layout.insertWidget(0, self.combo_box)

    def setup(self, wrappers: Iterable[Wrapper]):
        header = self.tr("Add wrapper")
        self.setWindowTitle(header)
        self.setSubtitle(header)
        for wrapper in wrappers:
            self.combo_box.addItem(f"{wrapper.name} ({wrapper.as_str})", wrapper.as_str)

    @Slot(int)
    def __on_index_changed(self, index: int):
        command = self.combo_box.itemData(index, Qt.ItemDataRole.UserRole)
        self.line_edit.setText(command)


class WrapperWidget(QFrame):
    # object: current, object: new
    update_wrapper = Signal(object, object)
    # object: current
    delete_wrapper = Signal(object)

    def __init__(self, wrapper: Wrapper, parent=None):
        super(WrapperWidget, self).__init__(parent=parent)
        self.setFrameShape(QFrame.Shape.StyledPanel)
        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Fixed)
        self.setToolTip(wrapper.as_str)

        text_lbl = QLabel(wrapper.name, parent=self)
        text_lbl.setFont(QFont("monospace"))
        text_lbl.setEnabled(wrapper.is_editable)

        image_lbl = QLabel(parent=self)
        image_lbl.setPixmap(qta_icon("mdi.drag-vertical").pixmap(QSize(20, 20)))

        edit_action = QAction("Edit", parent=self)
        edit_action.triggered.connect(self.__on_edit)
        delete_action = QAction("Delete", parent=self)
        delete_action.triggered.connect(self.__on_delete)

        manage_menu = QMenu(parent=self)
        manage_menu.addActions([edit_action, delete_action])

        manage_button = QPushButton(parent=self)
        manage_button.setIcon(qta_icon("mdi.menu", fallback="fa.align-justify"))
        manage_button.setMenu(manage_menu)
        manage_button.setEnabled(wrapper.is_editable)
        if not wrapper.is_editable:
            manage_button.setToolTip(self.tr("Manage through settings"))
        else:
            manage_button.setToolTip(self.tr("Manage"))

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(image_lbl)
        layout.addWidget(text_lbl)
        layout.addWidget(manage_button)
        self.setLayout(layout)

        self.wrapper = wrapper

        # lk: set object names for the stylesheet
        self.setObjectName(type(self).__name__)
        manage_button.setObjectName(f"{self.objectName()}Button")

    def data(self) -> Wrapper:
        return self.wrapper

    @Slot()
    def __on_delete(self) -> None:
        self.delete_wrapper.emit(self.wrapper)
        self.deleteLater()

    @Slot()
    def __on_edit(self) -> None:
        dialog = WrapperEditDialog(self)
        dialog.setup(self.wrapper)
        dialog.result_ready.connect(self.__on_edit_result)
        dialog.show()

    @Slot(bool, str)
    def __on_edit_result(self, accepted: bool, command: str):
        if accepted and command:
            new_wrapper = Wrapper(command=shlex.split(command))
            self.update_wrapper.emit(self.wrapper, new_wrapper)
            self.deleteLater()

    def mouseMoveEvent(self, a0: QMouseEvent) -> None:
        if a0.buttons() == Qt.MouseButton.LeftButton:
            a0.accept()
            if self.wrapper.is_compat_tool:
                return
            drag = QDrag(self)
            mime = QMimeData()
            drag.setMimeData(mime)
            drag.exec_(Qt.DropAction.MoveAction)


class WrapperSettingsScroll(QScrollArea):
    def __init__(self, parent=None):
        super(WrapperSettingsScroll, self).__init__(parent=parent)
        self.setFrameShape(QFrame.Shape.StyledPanel)
        self.setSizeAdjustPolicy(QScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setAlignment(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
        self.setWidgetResizable(True)
        self.setProperty("no_kinetic_scroll", True)

        self.setObjectName(type(self).__name__)
        self.horizontalScrollBar().setObjectName(f"{self.objectName()}Bar")
        self.verticalScrollBar().setObjectName(f"{self.objectName()}Bar")

    def setWidget(self, w):
        super().setWidget(w)
        w.installEventFilter(self)

    def eventFilter(self, a0: QObject, a1: QEvent) -> bool:
        if a0 is self.widget() and a1.type() == QEvent.Type.Resize:
            self.__resize(a0)
            return a0.event(a1)
        return False

    def __resize(self, e: QResizeEvent):
        minh = self.horizontalScrollBar().minimum()
        maxh = self.horizontalScrollBar().maximum()
        # lk: when the scrollbar is not visible, min and max are 0
        if maxh > minh:
            height = (
                e.size().height()
                + self.rect().height() // 2
                - self.contentsRect().height() // 2
                + self.widget().layout().spacing()
                + self.horizontalScrollBar().sizeHint().height()
            )
        else:
            height = e.size().height() + self.rect().height() - self.contentsRect().height()
        self.setMaximumHeight(max(height, self.minimumHeight()))


class WrapperSettings(QWidget):
    def __init__(self, parent=None):
        super(WrapperSettings, self).__init__(parent=parent)

        self.wrapper_label = QLabel(self.tr("No wrappers defined"), self)
        self.wrapper_label.setFrameStyle(QLabel.Shape.StyledPanel | QLabel.Shadow.Plain)
        self.wrapper_label.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)

        self.add_button = QPushButton(self.tr("Add wrapper"), self)
        self.add_button.clicked.connect(self.__on_add)

        self.wrapper_scroll = WrapperSettingsScroll(self)
        self.wrapper_scroll.setMinimumHeight(self.add_button.minimumSizeHint().height())

        self.wrapper_container = WrapperContainer(self.wrapper_label, self.wrapper_scroll)
        self.wrapper_container.orderChanged.connect(self.__on_order_changed)
        self.wrapper_scroll.setWidget(self.wrapper_container)

        # lk: set object names for the stylesheet
        self.setObjectName("WrapperSettings")
        self.wrapper_label.setObjectName(f"{self.objectName()}Label")

        main_layout = QHBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.addWidget(self.wrapper_scroll, alignment=Qt.AlignmentFlag.AlignTop)
        main_layout.addWidget(self.add_button, alignment=Qt.AlignmentFlag.AlignTop)

        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)

        self.app_name: str = "default"
        self.core = RareCore.instance().core()
        self.wrappers = RareCore.instance().wrappers()

    def showEvent(self, a0: QShowEvent):
        if a0.spontaneous():
            return super().showEvent(a0)
        self.update_state()
        return super().showEvent(a0)

    @Slot(QWidget, int)
    def __on_order_changed(self, widget: WrapperWidget, new_index: int):
        wrapper = widget.data()
        wrappers = self.wrappers.get_game_wrapper_list(self.app_name)
        wrappers.remove(wrapper)
        wrappers.insert(new_index, wrapper)
        self.wrappers.set_game_wrapper_list(self.app_name, wrappers)

    @Slot()
    def __on_add(self) -> None:
        dialog = WrapperAddDialog(self)
        dialog.setup(self.wrappers.user_wrappers)
        dialog.result_ready.connect(self.__on_add_result)
        dialog.show()

    @Slot(bool, str)
    def __on_add_result(self, accepted: bool, command: str):
        if accepted and command:
            wrapper = Wrapper(shlex.split(command))
            self.add_user_wrapper(wrapper)

    def __add_wrapper(self, wrapper: Wrapper, position: int = -1):
        self.wrapper_label.setVisible(False)
        widget = WrapperWidget(wrapper, self.wrapper_container)
        if position < 0:
            self.wrapper_container.addWidget(widget)
        else:
            self.wrapper_container.insertWidget(position, widget)
        widget.update_wrapper.connect(self.__update_wrapper)
        widget.delete_wrapper.connect(self.__delete_wrapper)

    def add_wrapper(self, wrapper: Wrapper, position: int = -1):
        wrappers = self.wrappers.get_game_wrapper_list(self.app_name)
        if position < 0 or wrapper.is_compat_tool:
            wrappers.append(wrapper)
        else:
            wrappers.insert(position, wrapper)
        self.wrappers.set_game_wrapper_list(self.app_name, wrappers)
        self.__add_wrapper(wrapper, position)

    def add_user_wrapper(self, wrapper: Wrapper, position: int = -1):
        if not wrapper:
            return

        if pf.system() in {"Linux", "FreeBSD"}:
            compat_cmds = [tool.command() for tool in steam.find_tools()]
            if wrapper.as_str in compat_cmds:
                QMessageBox.warning(
                    self,
                    self.tr("Warning"),
                    self.tr("Do not insert compatibility tools manually. Add them through Proton settings"),
                )
                return

        if wrapper.checksum in self.wrappers.get_game_md5sum_list(self.app_name):
            QMessageBox.warning(
                self,
                self.tr("Warning"),
                self.tr("Wrapper <b>{0}</b> is already in the list").format(wrapper.as_str),
            )
            return

        if not shutil.which(wrapper.executable):
            ans = QMessageBox.question(
                self,
                self.tr("Warning"),
                self.tr("Wrapper <b>{0}</b> is not in $PATH. Add it anyway?").format(wrapper.executable),
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )
            if ans == QMessageBox.StandardButton.No:
                return

        self.add_wrapper(wrapper, position)

    @Slot(object)
    def __delete_wrapper(self, wrapper: Wrapper):
        wrappers = self.wrappers.get_game_wrapper_list(self.app_name)
        wrappers.remove(wrapper)
        self.wrappers.set_game_wrapper_list(self.app_name, wrappers)
        if not wrappers:
            self.wrapper_label.setVisible(True)

    @Slot(object, object)
    def __update_wrapper(self, old: Wrapper, new: Wrapper):
        wrappers = self.wrappers.get_game_wrapper_list(self.app_name)
        index = wrappers.index(old)
        wrappers.remove(old)
        wrappers.insert(index, new)
        self.wrappers.set_game_wrapper_list(self.app_name, wrappers)
        self.__add_wrapper(new, index)

    @Slot()
    def update_state(self):
        for w in self.wrapper_container.findChildren(WrapperWidget, options=Qt.FindChildOption.FindDirectChildrenOnly):
            w.deleteLater()
        wrappers = self.wrappers.get_game_wrapper_list(self.app_name)
        if not wrappers:
            self.wrapper_label.setVisible(True)
        for wrapper in wrappers:
            self.__add_wrapper(wrapper)


class WrapperContainer(QWidget):
    # QWidget: moving widget, int: new index
    orderChanged: Signal = Signal(QWidget, int)

    def __init__(self, label: QLabel, parent=None):
        super(WrapperContainer, self).__init__(parent=parent)
        self.setAcceptDrops(True)
        self.__layout = QHBoxLayout()
        self.__drag_widget: Optional[QWidget] = None

        main_layout = QHBoxLayout(self)
        main_layout.addWidget(label)
        main_layout.addLayout(self.__layout)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setAlignment(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
        main_layout.setSizeConstraint(QHBoxLayout.SizeConstraint.SetFixedSize)

        # lk: set object names for the stylesheet
        self.setObjectName(type(self).__name__)

    # def count(self) -> int:
    #     return self.__layout.count()
    #
    # def itemData(self, index: int) -> Any:
    #     widget: WrapperWidget = self.__layout.itemAt(index).widget()
    #     return widget.data()

    def addWidget(self, widget: WrapperWidget):
        self.__layout.addWidget(widget)

    def insertWidget(self, index: int, widget: WrapperWidget):
        self.__layout.insertWidget(index, widget)

    def dragEnterEvent(self, e: QDragEnterEvent):
        widget = e.source()
        self.__drag_widget = widget
        e.accept()

    def __get_drop_index(self, x) -> int:
        drag_idx = self.__layout.indexOf(self.__drag_widget)

        if drag_idx > 0:
            prev_widget = self.__layout.itemAt(drag_idx - 1).widget()
            if x < self.__drag_widget.x() - prev_widget.width() // 2:
                return drag_idx - 1
        if drag_idx < self.__layout.count() - 1:
            next_widget = self.__layout.itemAt(drag_idx + 1).widget()
            if x > self.__drag_widget.x() + self.__drag_widget.width() + next_widget.width() // 2:
                return drag_idx + 1

        return drag_idx

    def dragMoveEvent(self, e: QDragMoveEvent) -> None:
        new_x = self.__get_drop_index(e.pos().x())
        self.__layout.insertWidget(new_x, self.__drag_widget)

    def dropEvent(self, e: QDropEvent):
        pos = e.pos()
        widget = e.source()
        new_x = self.__get_drop_index(pos.x())
        self.__layout.insertWidget(new_x, widget)
        self.__drag_widget = None
        self.orderChanged.emit(widget, new_x)
        e.accept()



from PySide6.QtCore import Signal
from PySide6.QtWidgets import QVBoxLayout, QGroupBox

from rare.models.game import RareGame
from rare.models.install import SelectiveDownloadsModel
from rare.utils.misc import qta_icon
from rare.widgets.dialogs import ButtonDialog, game_title
from rare.widgets.selective_widget import SelectiveWidget


class SelectiveDialog(ButtonDialog):
    result_ready = Signal(RareGame, SelectiveDownloadsModel)

    def __init__(self, rgame: RareGame, parent=None):
        super(SelectiveDialog, self).__init__(parent=parent)
        header = self.tr("Optional downloads for")
        self.setWindowTitle(game_title(header, rgame.app_title))
        self.setSubtitle(game_title(header, rgame.app_title))

        self.rgame = rgame
        self.selective_widget = SelectiveWidget(rgame, rgame.igame.platform, self)

        container = QGroupBox(self.tr("Optional downloads"), self)
        container_layout = QVBoxLayout(container)
        container_layout.setContentsMargins(0, 0, 0, 0)
        container_layout.addWidget(self.selective_widget)

        layout = QVBoxLayout()
        layout.addWidget(container)

        self.setCentralLayout(layout)

        self.accept_button.setText(self.tr("Verify"))
        self.accept_button.setIcon(qta_icon("fa.check"))

        self.options: SelectiveDownloadsModel = SelectiveDownloadsModel(rgame.app_name)

    def done_handler(self):
        self.result_ready.emit(self.rgame, self.options)

    def accept_handler(self):
        self.options.accepted = True
        self.options.install_tag = self.selective_widget.install_tags()

    def reject_handler(self):
        self.options.accepted = False
        self.options.install_tag = None



from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import List


@dataclass
class SearchDateRange:
    start_date: datetime = datetime(year=1990, month=1, day=1, tzinfo=timezone.utc)
    end_date: datetime = datetime.now(timezone.utc)

    def __str__(self):
        def fmt_date(date: datetime) -> str:
            # lk: The formatting accepted by the GraphQL API is either '%Y-%m-%dT%H:%M:%S.000Z' or '%Y-%m-%d'
            return datetime.strftime(date, '%Y-%m-%dT%H:%M:%S.000Z')
        return f"[{fmt_date(self.start_date)},{fmt_date(self.end_date)}]"


@dataclass
class SearchStoreQuery:
    country: str = "US"
    category: str = "games/edition/base|bundles/games|editors|software/edition/base"
    count: int = 30
    keywords: str = ""
    language: str = "en"
    namespace: str = ""
    with_mapping: bool = True
    item_ns: str = ""
    sort_by: str = "releaseDate"
    sort_dir: str = "DESC"
    start: int = 0
    tag: List[str] = ""
    release_date: SearchDateRange = field(default_factory=SearchDateRange)
    with_price: bool = True
    with_promotions: bool = True
    price_range: str = ""
    free_game: bool = None
    on_sale: bool = None
    effective_date: SearchDateRange = field(default_factory=SearchDateRange)

    def __post_init__(self):
        self.locale = f"{self.language}-{self.country}"

    def to_dict(self):
        payload = {
            "allowCountries": self.country,
            "category": self.category,
            "count": self.count,
            "country": self.country,
            "keywords": self.keywords,
            "locale": self.locale,
            "namespace": self.namespace,
            "withMapping": self.with_mapping,
            "itemNs": self.item_ns,
            "sortBy": self.sort_by,
            "sortDir": self.sort_dir,
            "start": self.start,
            "tag": self.tag,
            "releaseDate": str(self.release_date),
            "withPrice": self.with_price,
            "withPromotions": self.with_promotions,
            "priceRange": self.price_range,
            "freeGame": self.free_game,
            "onSale": self.on_sale,
            "effectiveDate": str(self.effective_date),
        }
        # payload.pop("withPromotions")
        payload.pop("onSale")
        if self.price_range == "free":
            payload["freeGame"] = True
            payload.pop("priceRange")
        elif self.price_range.startswith("<price>"):
            payload["priceRange"] = self.price_range.replace("<price>", "")
        if self.on_sale:
            payload["onSale"] = True

        if self.price_range:
            payload["effectiveDate"] = self.effective_date
        else:
            payload.pop("priceRange")
        return payload



import logging
from dataclasses import dataclass, field
from typing import List, Dict, Any, Type, Optional, Tuple

logger = logging.getLogger("DieselModels")

# lk: Typing overloads for unimplemented types
DieselSocialLinks = Dict


@dataclass
class DieselSystemDetailItem:
    _type: str = None
    minimum: str = None
    recommended: str = None
    title: str = None
    unmapped: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls: Type["DieselSystemDetailItem"], src: Dict[str, Any]) -> "DieselSystemDetailItem":
        d = src.copy()
        return cls(
            _type=d.pop("_type", ""),
            minimum=d.pop("minimum", ""),
            recommended=d.pop("recommended", ""),
            title=d.pop("title", ""),
            unmapped=d
        )


@dataclass
class DieselSystemDetail:
    _type: str = None
    details: Tuple[DieselSystemDetailItem, ...] = None
    systemType: str = None
    unmapped: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls: Type["DieselSystemDetail"], src: Dict[str, Any]) -> "DieselSystemDetail":
        d = src.copy()
        details = tuple(map(DieselSystemDetailItem.from_dict, d.pop("details", [])))
        return cls(
            _type=d.pop("_type", ""),
            details=details,
            systemType=d.pop("systemType", ""),
            unmapped=d
        )


@dataclass
class DieselSystemDetails:
    _type: str = None
    languages: List[str] = None
    rating: Dict = None
    systems: Tuple[DieselSystemDetail, ...] = None
    unmapped: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls: Type["DieselSystemDetails"], src: Dict[str, Any]) -> "DieselSystemDetails":
        d = src.copy()
        systems = tuple(map(DieselSystemDetail.from_dict, d.pop("systems", [])))
        return cls(
            _type=d.pop("_type", ""),
            languages=d.pop("languages", []),
            rating=d.pop("rating", {}),
            systems=systems,
            unmapped=d
        )


@dataclass
class DieselProductAbout:
    _type: str = None
    desciption: str = None
    developerAttribution: str = None
    publisherAttribution: str = None
    shortDescription: str = None
    unmapped: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls: Type["DieselProductAbout"], src: Dict[str, Any]) -> "DieselProductAbout":
        d = src.copy()
        return cls(
            _type=d.pop("_type", ""),
            desciption=d.pop("description", ""),
            developerAttribution=d.pop("developerAttribution", ""),
            publisherAttribution=d.pop("publisherAttribution", ""),
            shortDescription=d.pop("shortDescription", ""),
            unmapped=d
        )


@dataclass
class DieselProductDetail:
    _type: str = None
    about: DieselProductAbout = None
    requirements: DieselSystemDetails = None
    socialLinks: DieselSocialLinks = None
    unmapped: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls: Type["DieselProductDetail"], src: Dict[str, Any]) -> "DieselProductDetail":
        d = src.copy()
        about = DieselProductAbout.from_dict(x) if (x := d.pop("about"), {}) else None
        requirements = DieselSystemDetails.from_dict(x) if (x := d.pop("requirements", {})) else None
        return cls(
            _type=d.pop("_type", ""),
            about=about,
            requirements=requirements,
            socialLinks=d.pop("socialLinks", {}),
            unmapped=d
        )


@dataclass
class DieselProduct:
    _id: str = None
    _images_: List[str] = None
    _locale: str = None
    _slug: str = None
    _title: str = None
    _urlPattern: str = None
    namespace: str = None
    pages: Tuple["DieselProduct", ...] = None
    data: DieselProductDetail = None
    productName: str = None
    unmapped: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls: Type["DieselProduct"], src: Dict[str, Any]) -> "DieselProduct":
        d = src.copy()
        pages = tuple(map(DieselProduct.from_dict, d.pop("pages", [])))
        data = DieselProductDetail.from_dict(x) if (x := d.pop("data", {})) else None
        return cls(
            _id=d.pop("_id", ""),
            _images_=d.pop("_images_", []),
            _locale=d.pop("_locale", ""),
            _slug=d.pop("_slug", ""),
            _title=d.pop("_title", ""),
            _urlPattern=d.pop("_urlPattern", ""),
            namespace=d.pop("namespace", ""),
            pages=pages,
            data=data,
            productName=d.pop("productName", ""),
            unmapped=d
        )



import os
import platform
from datetime import datetime
from logging import getLogger
from typing import Tuple

from PySide6.QtCore import QThreadPool, QSettings, Slot
from PySide6.QtWidgets import (
    QWidget,
    QFileDialog,
    QLabel,
    QPushButton,
    QSizePolicy,
    QMessageBox,
    QGroupBox,
    QVBoxLayout,
    QSpacerItem, QFormLayout,
)
from legendary.models.game import SaveGameStatus

from rare.models.game import RareGame
from rare.models.options import options
from rare.shared import RareCore
from rare.shared.workers.wine_resolver import WineSavePathResolver
from rare.ui.components.tabs.games.game_info.cloud_settings_widget import Ui_CloudSettingsWidget
from rare.ui.components.tabs.games.game_info.cloud_sync_widget import Ui_CloudSyncWidget
from rare.utils.metrics import timelogger
from rare.utils.misc import qta_icon
from rare.widgets.indicator_edit import PathEdit, IndicatorReasonsCommon
from rare.widgets.loading_widget import LoadingWidget
from rare.widgets.side_tab import SideTabContents

logger = getLogger("CloudSaves")


class CloudSaves(QWidget, SideTabContents):
    def __init__(self, parent=None):
        super(CloudSaves, self).__init__(parent=parent)

        self.sync_widget = QWidget(self)
        self.sync_ui = Ui_CloudSyncWidget()
        self.sync_ui.setupUi(self.sync_widget)

        self.info_label = QLabel(self.tr("<b>This game doesn't support cloud saves</b>"))

        self.rcore = RareCore.instance()
        self.core = RareCore.instance().core()
        self.settings = QSettings()

        self.sync_ui.icon_local.setPixmap(qta_icon("mdi.harddisk", "fa.desktop").pixmap(128, 128))
        self.sync_ui.icon_remote.setPixmap(qta_icon("mdi.cloud-outline", "ei.cloud").pixmap(128, 128))

        self.sync_ui.upload_button.clicked.connect(self.upload)
        self.sync_ui.download_button.clicked.connect(self.download)

        self.loading_widget = LoadingWidget(parent=self.sync_widget)
        self.loading_widget.setVisible(False)

        self.rgame: RareGame = None

        self.cloud_widget = QGroupBox(self)
        self.cloud_ui = Ui_CloudSettingsWidget()
        self.cloud_ui.setupUi(self.cloud_widget)

        self.cloud_save_path_edit = PathEdit(
            path="",
            file_mode=QFileDialog.FileMode.Directory,
            placeholder=self.tr('Use "Calculate path" or "Browse" ...'),
            edit_func=self.edit_save_path,
            save_func=self.save_save_path,
        )
        self.cloud_ui.main_layout.setWidget(
            self.cloud_ui.main_layout.getWidgetPosition(self.cloud_ui.path_label)[0],
            QFormLayout.ItemRole.FieldRole,
            self.cloud_save_path_edit
        )

        self.compute_save_path_button = QPushButton(qta_icon("fa.magic"), self.tr("Calculate path"))
        self.compute_save_path_button.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Fixed)
        self.compute_save_path_button.clicked.connect(self.compute_save_path)
        self.cloud_ui.main_layout.addRow(None, self.compute_save_path_button)

        self.cloud_ui.sync_check.stateChanged.connect(self.__on_sync_check_changed)

        layout = QVBoxLayout(self)
        layout.addWidget(self.sync_widget)
        layout.addWidget(self.cloud_widget)
        layout.addWidget(self.info_label)
        layout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Expanding))

    @staticmethod
    def edit_save_path(text: str) -> Tuple[bool, str, int]:
        if platform.system() != "Windows":
            if os.path.exists(text):
                return True, text, IndicatorReasonsCommon.VALID
            else:
                return False, text, IndicatorReasonsCommon.DIR_NOT_EXISTS
        return True, text, IndicatorReasonsCommon.VALID

    def save_save_path(self, text: str):
        if text != self.rgame.save_path:
            self.rgame.save_path = text

    def upload(self):
        self.sync_widget.setEnabled(False)
        self.loading_widget.setVisible(True)
        self.rgame.upload_saves()

    def download(self):
        self.sync_widget.setEnabled(False)
        self.loading_widget.setVisible(True)
        self.rgame.download_saves()

    def compute_save_path(self):
        if self.rgame.is_installed and self.rgame.game.supports_cloud_saves:
            try:
                with timelogger(logger, "Detecting save path"):
                    new_path = self.core.get_save_path(self.rgame.app_name)
                if platform.system() != "Windows" and not os.path.exists(new_path):
                    raise ValueError(f'Path "{new_path}" does not exist.')
            except Exception as e:
                logger.warning(str(e))
                resolver = WineSavePathResolver(self.core, self.rgame)
                # if not resolver.environ.get("WINEPREFIX"):
                #     del resolver
                #     self.cloud_save_path_edit.setText("")
                #     QMessageBox.warning(self, "Warning", "No wine prefix selected. Please set it in settings")
                #     return
                self.cloud_save_path_edit.setText(self.tr("Loading..."))
                self.cloud_save_path_edit.setDisabled(True)
                self.compute_save_path_button.setDisabled(True)

                resolver.signals.result_ready.connect(self.__on_wine_resolver_result)
                QThreadPool.globalInstance().start(resolver)
                return
            else:
                self.cloud_save_path_edit.setText(new_path)

    @Slot()
    def __on_sync_check_changed(self):
        if self.settings.value(*options.auto_sync_cloud) == self.cloud_ui.sync_check.isChecked():
            self.settings.remove(f"{self.rgame.app_name}/{options.auto_sync_cloud.key}")
        else:
            self.settings.setValue(f"{self.rgame.app_name}/auto_sync_cloud", self.cloud_ui.sync_check.isChecked())

    @Slot(str, str)
    def __on_wine_resolver_result(self, path, app_name):
        logger.info("Wine resolver finished for %s", app_name)
        logger.info("Computed save path: %s", path)
        if app_name == self.rgame.app_name:
            self.cloud_save_path_edit.setDisabled(False)
            self.compute_save_path_button.setDisabled(False)
            if path and not os.path.exists(path):
                try:
                    os.makedirs(path, exist_ok=True)
                except PermissionError:
                    self.cloud_save_path_edit.setText("")
                    QMessageBox.warning(
                        self,
                        self.tr("Error - {}").format(self.rgame.app_title),
                        self.tr(
                            "Error while calculating path for <b>{}</b>. Insufficient permissions to create <b>{}</b>"
                        ).format(self.rgame.app_title, path),
                    )
                    return
            if not path:
                self.cloud_save_path_edit.setText("")
                return
            self.cloud_save_path_edit.setText(path)

    def __update_widget(self):
        supports_saves = self.rgame.igame is not None and (
                self.rgame.game.supports_cloud_saves or self.rgame.game.supports_mac_cloud_saves
        )

        self.sync_widget.setEnabled(
            bool(supports_saves and self.rgame.save_path))  # and not self.rgame.is_save_up_to_date))

        self.cloud_widget.setEnabled(supports_saves)
        self.info_label.setVisible(not supports_saves)
        if not supports_saves:
            self.sync_ui.date_info_local.setText("None")
            self.sync_ui.age_label_local.setText("None")
            self.sync_ui.date_info_remote.setText("None")
            self.sync_ui.age_label_remote.setText("None")
            self.cloud_ui.sync_check.setChecked(False)
            self.cloud_save_path_edit.setText("")
            return

        status, (dt_local, dt_remote) = self.rgame.save_game_state

        local_tz = datetime.now().astimezone().tzinfo
        self.sync_ui.date_info_local.setText(
            dt_local.astimezone(local_tz).strftime("%A, %d %B %Y %X") if dt_local else "None")
        self.sync_ui.date_info_remote.setText(
            dt_remote.astimezone(local_tz).strftime("%A, %d %B %Y %X") if dt_remote else "None")

        newer = self.tr("Newer")
        self.sync_ui.age_label_local.setText(
            f"<b>{newer}</b>" if status == SaveGameStatus.LOCAL_NEWER else " "
        )
        self.sync_ui.age_label_remote.setText(
            f"<b>{newer}</b>" if status == SaveGameStatus.REMOTE_NEWER else " "
        )

        button_disabled = self.rgame.state in [RareGame.State.RUNNING, RareGame.State.SYNCING]
        self.sync_widget.setDisabled(button_disabled)
        if self.rgame.state == RareGame.State.SYNCING:
            self.loading_widget.start()
        else:
            self.loading_widget.stop()

        self.sync_ui.upload_button.setDisabled(not dt_local)
        self.sync_ui.download_button.setDisabled(not dt_remote)

        self.cloud_ui.sync_check.blockSignals(True)
        self.cloud_ui.sync_check.setChecked(self.rgame.auto_sync_saves)
        self.cloud_ui.sync_check.blockSignals(False)

        self.cloud_save_path_edit.setText(self.rgame.save_path if self.rgame.save_path else "")
        if platform.system() == "Windows" and not self.rgame.save_path:
            self.compute_save_path()

    def update_game(self, rgame: RareGame):
        if self.rgame:
            self.rgame.signals.widget.update.disconnect(self.__update_widget)

        self.rgame = rgame

        self.set_title.emit(rgame.app_title)
        rgame.signals.widget.update.connect(self.__update_widget)

        self.__update_widget()



import locale
import platform as pf
from argparse import Namespace
from typing import Any, Type
from .library import LibraryFilter, LibraryOrder, LibraryView


class Value(Namespace):
    key: str
    default: Any
    dtype: Type

    def __len__(self):
        return len(self.__dict__)

    def __iter__(self):
        yield from self.__dict__.values()


# They key names are set to the existing option name
class Defaults(Namespace):
    win32_meta = Value(key="win32_meta", default=False, dtype=bool)
    macos_meta = Value(key="macos_meta", default=pf.system() == "Darwin", dtype=bool)
    unreal_meta = Value(key="unreal_meta", default=False, dtype=bool)
    exclude_non_asset = Value(key="exclude_non_asset", default=False, dtype=bool)
    exclude_entitlements = Value(key="exclude_entitlements", default=False, dtype=bool)

    language = Value(key="language", default=locale.getlocale()[0], dtype=str)
    sys_tray = Value(key="sys_tray", default=True, dtype=bool)
    auto_update = Value(key="auto_update", default=False, dtype=bool)
    auto_sync_cloud = Value(key="auto_sync_cloud", default=False, dtype=bool)
    confirm_start = Value(key="confirm_start", default=False, dtype=bool)
    restore_window = Value(key="save_size", default=False, dtype=bool)
    window_width = Value(key="window_width", default=1280, dtype=int)
    window_height = Value(key="window_height", default=720, dtype=int)
    notification = Value(key="notification", default=True, dtype=bool)
    log_games = Value(key="show_console", default=False, dtype=bool)

    color_scheme = Value(key="color_scheme", default="", dtype=str)
    style_sheet = Value(key="style_sheet", default="RareStyle", dtype=str)

    library_view = Value(key="library_view", default=int(LibraryView.COVER), dtype=int)
    library_filter = Value(
        key="library_filter",
        default=int(LibraryFilter.MAC if pf.system() == "Darwin" else LibraryFilter.ALL), dtype=int
    )
    library_order = Value(key="library_order", default=int(LibraryOrder.TITLE), dtype=int)

    discord_rpc_mode = Value(key="discord_rpc_mode", default=0, dtype=int)
    discord_rpc_game = Value(key="discord_rpc_game", default=True, dtype=bool)
    discord_rpc_time = Value(key="discord_rpc_time", default=True, dtype=bool)
    discord_rpc_os = Value(key="discord_rpc_os", default=True, dtype=bool)


options = Defaults()

__all__ = ['options', 'LibraryFilter', 'LibraryOrder', 'LibraryView']



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'overlay.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QFormLayout, QGridLayout,
    QGroupBox, QLabel, QSizePolicy, QSpacerItem,
    QWidget)

class Ui_OverlaySettings(object):
    def setupUi(self, OverlaySettings):
        if not OverlaySettings.objectName():
            OverlaySettings.setObjectName(u"OverlaySettings")
        OverlaySettings.resize(555, 188)
        OverlaySettings.setWindowTitle(u"OverlaySettings")
        OverlaySettings.setTitle(u"")
        self.main_layout = QGridLayout(OverlaySettings)
        self.main_layout.setObjectName(u"main_layout")
        self.show_overlay_label = QLabel(OverlaySettings)
        self.show_overlay_label.setObjectName(u"show_overlay_label")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.show_overlay_label.sizePolicy().hasHeightForWidth())
        self.show_overlay_label.setSizePolicy(sizePolicy)

        self.main_layout.addWidget(self.show_overlay_label, 0, 0, 1, 1)

        self.show_overlay_combo = QComboBox(OverlaySettings)
        self.show_overlay_combo.setObjectName(u"show_overlay_combo")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.show_overlay_combo.sizePolicy().hasHeightForWidth())
        self.show_overlay_combo.setSizePolicy(sizePolicy1)

        self.main_layout.addWidget(self.show_overlay_combo, 0, 1, 1, 2)

        self.options_group = QGroupBox(OverlaySettings)
        self.options_group.setObjectName(u"options_group")
        self.options_group.setTitle(u"")
        self.options_layout = QGridLayout(self.options_group)
        self.options_layout.setObjectName(u"options_layout")
        self.options_form = QFormLayout()
        self.options_form.setObjectName(u"options_form")
        self.options_form.setLabelAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.options_layout.addLayout(self.options_form, 1, 0, 1, 1)

        self.options_grid = QGridLayout()
        self.options_grid.setObjectName(u"options_grid")

        self.options_layout.addLayout(self.options_grid, 0, 0, 1, 1)

        self.hspacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.options_layout.addItem(self.hspacer, 0, 1, 2, 1)


        self.main_layout.addWidget(self.options_group, 2, 0, 1, 3)


        self.retranslateUi(OverlaySettings)
    # setupUi

    def retranslateUi(self, OverlaySettings):
        self.show_overlay_label.setText(QCoreApplication.translate("OverlaySettings", u"Show HUD", None))
        pass
    # retranslateUi




import sys
from datetime import datetime
from enum import IntEnum
from logging import getLogger

from PySide6.QtCore import Signal, Slot
from PySide6.QtWidgets import QApplication, QWidget, QVBoxLayout, QLabel, QDialog
from legendary.core import LegendaryCore
from legendary.models.game import InstalledGame

from rare.ui.components.tabs.games.game_info.cloud_sync_widget import Ui_CloudSyncWidget
from rare.utils.misc import qta_icon
from rare.widgets.dialogs import ButtonDialog, game_title

logger = getLogger("CloudSyncDialog")


class CloudSyncDialogResult(IntEnum):
    DOWNLOAD = 2
    UPLOAD = 1
    CANCEL = 0
    SKIP = 3


class CloudSyncDialog(ButtonDialog):
    result_ready: Signal = Signal(CloudSyncDialogResult)

    def __init__(self, igame: InstalledGame, dt_local: datetime, dt_remote: datetime, parent=None):
        super(CloudSyncDialog, self).__init__(parent=parent)
        header = self.tr("Cloud saves for")
        self.setWindowTitle(game_title(header, igame.title))

        title_label = QLabel(f"<h4>{game_title(header, igame.title)}</h4>", self)

        sync_widget = QWidget(self)
        self.sync_ui = Ui_CloudSyncWidget()
        self.sync_ui.setupUi(sync_widget)

        layout = QVBoxLayout()
        layout.addWidget(title_label)
        layout.addWidget(sync_widget)

        self.accept_button.setText(self.tr("Skip"))
        self.accept_button.setIcon(qta_icon("fa.chevron-right"))

        self.setCentralLayout(layout)

        self.status = CloudSyncDialogResult.CANCEL

        newer = self.tr("Newer")
        if dt_remote and dt_local:
            self.sync_ui.age_label_local.setText(f"<b>{newer}</b>" if dt_remote < dt_local else " ")
            self.sync_ui.age_label_remote.setText(f"<b>{newer}</b>" if dt_remote > dt_local else " ")
        # Set status, if one of them is None
        elif dt_remote and not dt_local:
            self.status = CloudSyncDialogResult.DOWNLOAD
        elif not dt_remote and dt_local:
            self.status = CloudSyncDialogResult.UPLOAD
        else:
            self.status = CloudSyncDialogResult.SKIP

        local_tz = datetime.now().astimezone().tzinfo
        self.sync_ui.date_info_local.setText(
            dt_local.astimezone(local_tz).strftime("%A, %d %B %Y %X") if dt_local else "None")
        self.sync_ui.date_info_remote.setText(
            dt_remote.astimezone(local_tz).strftime("%A, %d %B %Y %X") if dt_remote else "None")

        self.sync_ui.icon_local.setPixmap(qta_icon("mdi.harddisk", "fa.desktop").pixmap(128, 128))
        self.sync_ui.icon_remote.setPixmap(qta_icon("mdi.cloud-outline", "ei.cloud").pixmap(128, 128))

        self.sync_ui.upload_button.clicked.connect(self.__on_upload)
        self.sync_ui.download_button.clicked.connect(self.__on_download)

        if self.status == CloudSyncDialogResult.SKIP:
            self.accept()

    def __on_upload(self):
        self.status = CloudSyncDialogResult.UPLOAD
        self.done(QDialog.DialogCode.Accepted)

    def __on_download(self):
        self.status = CloudSyncDialogResult.DOWNLOAD
        self.done(QDialog.DialogCode.Accepted)

    def done_handler(self):
        self.result_ready.emit(self.status)

    def accept_handler(self):
        self.status = CloudSyncDialogResult.SKIP

    def reject_handler(self):
        self.status = CloudSyncDialogResult.CANCEL


if __name__ == "__main__":
    app = QApplication(sys.argv)
    core = LegendaryCore()

    @Slot(int)
    def __callback(status: int):
        print(repr(CloudSyncDialogResult(status)))

    dlg = CloudSyncDialog(core.get_installed_list()[0], datetime.now(), datetime.strptime("2021,1", "%Y,%M"))
    dlg.result_ready.connect(__callback)
    dlg.open()
    app.exec()



from PySide6.QtCore import QSettings
from PySide6.QtWidgets import QGroupBox

from rare.shared import GlobalSignalsSingleton
from rare.models.options import options
from rare.ui.components.tabs.settings.widgets.discord_rpc import Ui_DiscordRPCSettings


class DiscordRPCSettings(QGroupBox):
    def __init__(self, parent):
        super(DiscordRPCSettings, self).__init__(parent=parent)
        self.ui = Ui_DiscordRPCSettings()
        self.ui.setupUi(self)
        self.signals = GlobalSignalsSingleton()

        self.settings = QSettings()

        self.ui.enable.setCurrentIndex(self.settings.value(*options.discord_rpc_mode))
        self.ui.enable.currentIndexChanged.connect(self.__enable_changed)

        self.ui.show_game.setChecked((self.settings.value(*options.discord_rpc_game)))
        self.ui.show_game.stateChanged.connect(
            lambda: self.settings.setValue(options.discord_rpc_game.key, self.ui.show_game.isChecked())
        )

        self.ui.show_os.setChecked((self.settings.value(*options.discord_rpc_os)))
        self.ui.show_os.stateChanged.connect(
            lambda: self.settings.setValue(options.discord_rpc_os.key, self.ui.show_os.isChecked())
        )

        self.ui.show_time.setChecked((self.settings.value(*options.discord_rpc_time)))
        self.ui.show_time.stateChanged.connect(
            lambda: self.settings.setValue(options.discord_rpc_time.key, self.ui.show_time.isChecked())
        )

        try:
            import pypresence
        except ModuleNotFoundError:
            self.setDisabled(True)
            self.setToolTip(self.tr("Pypresence is not installed"))

    def __enable_changed(self, i):
        self.settings.setValue(options.discord_rpc_mode.key, i)
        self.signals.discord_rpc.update_settings.emit()



import os
from typing import Union, List

from legendary.core import LegendaryCore
from legendary.models.game import InstalledGame

from rare.utils.config_helper import get_prefixes


class PathSpec:

    @staticmethod
    def egl_appdata() -> str:
        return r"%LOCALAPPDATA%\EpicGamesLauncher\Saved\Config\Windows"

    @staticmethod
    def egl_programdata() -> str:
        return r"%PROGRAMDATA%\Epic\EpicGamesLauncher\Data\Manifests"

    @staticmethod
    def wine_programdata() -> str:
        return r"ProgramData"

    @staticmethod
    def wine_egl_programdata() -> str:
        return PathSpec.egl_programdata(
        ).replace(
            "\\", "/"
        ).replace(
            "%PROGRAMDATA%", PathSpec.wine_programdata()
        )

    @staticmethod
    def prefix_egl_programdata(prefix: str) -> str:
        return os.path.join(prefix, "dosdevices/c:", PathSpec.wine_egl_programdata())

    @staticmethod
    def wine_egl_prefixes(results: int = 0) -> Union[List[str], str]:
        possible_prefixes = get_prefixes()
        prefixes = [
            prefix
            for prefix, _ in possible_prefixes
            if os.path.exists(os.path.join(prefix, PathSpec.wine_egl_programdata()))
        ]
        if not prefixes:
            return ""
        if not results:
            return prefixes
        elif results == 1:
            return prefixes[0]
        else:
            return prefixes[:results]

    def __init__(self, core: LegendaryCore = None, igame: InstalledGame = None):
        self.__egl_path_vars = {
            "{appdata}": os.path.expandvars("%LOCALAPPDATA%"),
            "{userdir}": os.path.expandvars("%USERPROFILE%/Documents"),
            "{userprofile}": os.path.expandvars("%userprofile%"),  # possibly wrong
            "{usersavedgames}": os.path.expandvars("%USERPROFILE%/Saved Games"),
        }

        if core is not None:
            self.__egl_path_vars["{epicid}"] = core.lgd.userdata["account_id"]

        if igame is not None:
            self.__egl_path_vars["{installdir}"] = igame.install_path

    def resolve_egl_path_vars(self, path: str) -> Union[str, bytes]:
        cooked_path = (self.__egl_path_vars.get(p.lower(), p) for p in path.split("/"))
        return os.path.join(*cooked_path)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'dlc_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFormLayout, QFrame, QHBoxLayout,
    QLabel, QPushButton, QSizePolicy, QSpacerItem,
    QWidget)

class Ui_GameDlcWidget(object):
    def setupUi(self, GameDlcWidget):
        if not GameDlcWidget.objectName():
            GameDlcWidget.setObjectName(u"GameDlcWidget")
        GameDlcWidget.resize(454, 58)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(GameDlcWidget.sizePolicy().hasHeightForWidth())
        GameDlcWidget.setSizePolicy(sizePolicy)
        GameDlcWidget.setWindowTitle(u"GameDlcWidget")
        GameDlcWidget.setFrameShape(QFrame.Shape.StyledPanel)
        GameDlcWidget.setFrameShadow(QFrame.Shadow.Plain)
        self.dlc_layout = QHBoxLayout(GameDlcWidget)
        self.dlc_layout.setObjectName(u"dlc_layout")
        self.dlc_layout.setContentsMargins(0, 0, 0, 0)
        self.dlc_info = QWidget(GameDlcWidget)
        self.dlc_info.setObjectName(u"dlc_info")
        self.dlc_info_layout = QFormLayout(self.dlc_info)
        self.dlc_info_layout.setObjectName(u"dlc_info_layout")
        self.dlc_info_layout.setLabelAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.dlc_info_layout.setContentsMargins(0, 0, 0, 0)
        self.dlc_name_label = QLabel(self.dlc_info)
        self.dlc_name_label.setObjectName(u"dlc_name_label")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.dlc_name_label.sizePolicy().hasHeightForWidth())
        self.dlc_name_label.setSizePolicy(sizePolicy1)
        font = QFont()
        font.setBold(True)
        self.dlc_name_label.setFont(font)

        self.dlc_info_layout.setWidget(0, QFormLayout.LabelRole, self.dlc_name_label)

        self.dlc_name = QLabel(self.dlc_info)
        self.dlc_name.setObjectName(u"dlc_name")
        self.dlc_name.setMinimumSize(QSize(160, 0))
        self.dlc_name.setText(u"dlc_name")
        self.dlc_name.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

        self.dlc_info_layout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.dlc_name)

        self.version_label = QLabel(self.dlc_info)
        self.version_label.setObjectName(u"version_label")
        sizePolicy1.setHeightForWidth(self.version_label.sizePolicy().hasHeightForWidth())
        self.version_label.setSizePolicy(sizePolicy1)
        self.version_label.setFont(font)

        self.dlc_info_layout.setWidget(1, QFormLayout.LabelRole, self.version_label)

        self.version = QLabel(self.dlc_info)
        self.version.setObjectName(u"version")
        self.version.setMinimumSize(QSize(160, 0))
        self.version.setText(u"version")
        self.version.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

        self.dlc_info_layout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.version)

        self.app_name_label = QLabel(self.dlc_info)
        self.app_name_label.setObjectName(u"app_name_label")
        sizePolicy1.setHeightForWidth(self.app_name_label.sizePolicy().hasHeightForWidth())
        self.app_name_label.setSizePolicy(sizePolicy1)
        self.app_name_label.setFont(font)

        self.dlc_info_layout.setWidget(2, QFormLayout.LabelRole, self.app_name_label)

        self.app_name = QLabel(self.dlc_info)
        self.app_name.setObjectName(u"app_name")
        self.app_name.setMinimumSize(QSize(160, 0))
        self.app_name.setText(u"app_name")
        self.app_name.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

        self.dlc_info_layout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.app_name)


        self.dlc_layout.addWidget(self.dlc_info)

        self.hspacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.dlc_layout.addItem(self.hspacer)

        self.action_button = QPushButton(GameDlcWidget)
        self.action_button.setObjectName(u"action_button")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.action_button.sizePolicy().hasHeightForWidth())
        self.action_button.setSizePolicy(sizePolicy2)
        self.action_button.setMinimumSize(QSize(160, 0))
        self.action_button.setText(u"Action")

        self.dlc_layout.addWidget(self.action_button, 0, Qt.AlignmentFlag.AlignBottom)

        self.dlc_layout.setStretch(1, 1)

        self.retranslateUi(GameDlcWidget)
    # setupUi

    def retranslateUi(self, GameDlcWidget):
        self.dlc_name_label.setText(QCoreApplication.translate("GameDlcWidget", u"DLC Name", None))
        self.version_label.setText(QCoreApplication.translate("GameDlcWidget", u"Version", None))
        self.app_name_label.setText(QCoreApplication.translate("GameDlcWidget", u"Application", None))
        pass
    # retranslateUi




import os
import sys
from typing import Union, Type

import qstylizer.style
from PySide6.QtCore import QDir, QObject
from PySide6.QtGui import QColor
from PySide6.scripts.pyside_tool import qt_tool_wrapper
from shiboken6.Shiboken import Object as wrappertype

from rare.utils.misc import widget_object_name

verbose = True
compressLevel = 6
compressThreshold = 0


def css_name(widget: Union[wrappertype, QObject, Type], subwidget: str = ""):
    return f"#{widget_object_name(widget, '')}{subwidget}"


css = qstylizer.style.StyleSheet()


# Generic flat button
css['QPushButton[flat="true"]'].setValues(
    border="0px",
    borderRadius="5px",
    backgroundColor="rgba(255, 255, 255, 5%)",
)


# InfoLabel
css.QLabel["#InfoLabel"].setValues(
    color="#999",
    fontStyle="italic",
    fontWeight="normal",
)

# [Un]InstallButton
css.QPushButton["#InstallButton"].setValues(
    borderColor=QColor(0, 180, 0).name(),
    backgroundColor=QColor(0, 120, 0).name()
)
css.QPushButton["#InstallButton"].hover.setValues(
    borderColor=QColor(0, 135, 0).name(),
    backgroundColor=QColor(0, 90, 0).name()
)
css.QPushButton["#InstallButton"].disabled.setValues(
    borderColor=QColor(0, 60, 0).name(),
    backgroundColor=QColor(0, 40, 0).name()
)
css.QPushButton["#UninstallButton"].setValues(
    borderColor=QColor(180, 0, 0).name(),
    backgroundColor=QColor(120, 0, 0).name()
)
css.QPushButton["#UninstallButton"].hover.setValues(
    borderColor=QColor(135, 0, 0).name(),
    backgroundColor=QColor(90, 0, 0).name()
)
css.QPushButton["#UninstallButton"].disabled.setValues(
    borderColor=QColor(60, 0, 0).name(),
    backgroundColor=QColor(40, 0, 0).name()
)


# QueueWorkerLabel
css.QLabel["#QueueWorkerLabel"].setValues(
    borderWidth="1px",
    borderStyle="solid",
    borderRadius="3px",
)
verify_color = QColor("#d6af57")
css.QLabel["#QueueWorkerLabel"]['[workerType="Verify"]'].setValues(
    borderColor=verify_color.darker(200).name(),
    backgroundColor=verify_color.darker(400).name()
)
move_color = QColor("#41cad9")
css.QLabel["#QueueWorkerLabel"]['[workerType="Move"]'].setValues(
    borderColor=move_color.darker(200).name(),
    backgroundColor=move_color.darker(400).name()
)


# ProgressLabel
from rare.components.tabs.library.widgets.library_widget import ProgressLabel
css.QLabel[css_name(ProgressLabel)].setValues(
    borderWidth="1px",
    borderRadius="5%",
    fontWeight="bold",
    fontSize="16pt",
)


# IconGameWidget
from rare.components.tabs.library.widgets.icon_widget import IconWidget
icon_background_props = {
    "backgroundColor": "rgba(0, 0, 0, 65%)",
}
css.QLabel[css_name(IconWidget, "StatusLabel")].setValues(
    fontWeight="bold",
    color="white",
    **icon_background_props,
    borderRadius="5%",
    borderTopLeftRadius="11%",
    borderTopRightRadius="11%",
)
css.QWidget[css_name(IconWidget, "MiniWidget")].setValues(
    color="rgb(238, 238, 238)",
    **icon_background_props,
    borderRadius="5%",
    borderBottomLeftRadius="9%",
    borderBottomRightRadius="9%",
)
icon_bottom_label_props = {
    "color": "white",
    "backgroundColor": "rgba(0, 0, 0, 0%)",
}
css.QLabel[css_name(IconWidget, "TitleLabel")].setValues(
    fontWeight="bold",
    **icon_bottom_label_props,
)
css.QLabel[css_name(IconWidget, "TooltipLabel")].setValues(
    **icon_bottom_label_props,
)
icon_square_button_props = {
    "border": "1px solid black",
    "borderRadius": "10%",
}
icon_square_button_props.update(icon_background_props)
css.QPushButton[css_name(IconWidget, "Button")].setValues(
    **icon_square_button_props
)
css.QPushButton[css_name(IconWidget, "Button")].hover.borderColor.setValue("gray")


# ListGameWidget
from rare.components.tabs.library.widgets.list_widget import ListWidget
css.QLabel[css_name(ListWidget, "TitleLabel")].fontWeight.setValue("bold")
list_status_label_props = {
    "color": "white",
    "backgroundColor": "rgba(0, 0, 0, 75%)",
    "border": "1px solid black",
    "borderRadius": "5px",
}
css.QLabel[css_name(ListWidget, "StatusLabel")].setValues(
    fontWeight="bold",
    **list_status_label_props,
)
css.QLabel[css_name(ListWidget, "TooltipLabel")].setValues(
    **list_status_label_props,
)
css.QPushButton[css_name(ListWidget, "Button")].textAlign.setValue("left")
css.QLabel[css_name(ListWidget, "InfoLabel")].color.setValue("#999")


# SelectViewWidget
from rare.components.tabs.library.head_bar import SelectViewWidget
css.QPushButton[css_name(SelectViewWidget, "Button")].setValues(
    border="none",
    backgroundColor="transparent",
)


# ButtonLineEdit
from rare.widgets.button_edit import ButtonLineEdit
css.QPushButton[css_name(ButtonLineEdit, "Button")].setValues(
    backgroundColor="transparent",
    border="0px",
    padding="0px",
)


if __name__ == "__main__":
    with open("stylesheet.qss", "w", encoding="utf-8") as qss:
        qss.write(f'\n/* This file is auto-generated from "{os.path.basename(__file__)}". DO NOT EDIT!!! */\n\n')
        qss.write(css.toString())

    qt_tool_wrapper(
        "rcc",
        [
            "-g", "python",
            "--compress", str(compressLevel),
            "--threshold", str(compressThreshold),
            "--verbose" if verbose else "",
            "stylesheet.qrc",
            "-o", "__init__.py",
        ],
        True
    )



import sys

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QDialog, QApplication, QVBoxLayout
from legendary.core import LegendaryCore

from . import StoreTab


class StoreWindow(QDialog):
    def __init__(self):
        super().__init__()

        self.core = LegendaryCore()
        self.core.login()
        self.store_tab = StoreTab(self.core, self)

        layout = QVBoxLayout(self)
        layout.addWidget(self.store_tab)

        self.store_tab.show()


if __name__ == "__main__":
    import rare.resources.static_css
    # import rare.resources.stylesheets.RareStyle
    from rare.utils.misc import set_style_sheet

    app = QApplication(sys.argv)
    app.setApplicationName("Rare")
    app.setOrganizationName("Rare")

    set_style_sheet("")
    set_style_sheet("RareStyle")
    window = StoreWindow()
    window.setWindowTitle(f"{app.applicationName()} - Store")
    window.resize(QSize(1280, 800))
    window.show()
    app.exec()



# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 6.7.2
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore

qt_resource_data = b"\
\x00\x00\x02\xa9\
\x00\
\x00\x0a\xb9x\xda\xb5V]O\xa3@\x14}\xe7W\x8c\
5M\xd4X\x05\xba\xb4\xb6\xc6\x97\x8d>\x98\xe8\xaa\xb1\
\xab\x0ff\x1f\x06\xb8\xd0\x89\xe3\x0c\x19.[\xbb\x1b\xff\
\xfb\x0e\x85m\xa1\x0b\xd3\xee\xae6M\xa1\x03s\xce=\
\xe7~\x80u|@&S\x96\x92\x88q \xfaH3\
\x94\xbd\x18\x04(\x8a\x10\x92H\xc9\x17\xd2Iq\xce!\
\x9d\x02\xe0Q2\xef\x1c\x91\xf3\x1b\xf2\xe5fB.\xce\
/';;;\xe4\xe0\xd8\xb2\xeen\xb3t\xfa9C\
\x94\xe2)\xe2\x14\xcf:\xa82\xe8|#?-\xa2?\
\xbeT!\xa81\xb1\x93\xd7\xd3\xcaBO\xd1\x90e\xe9\
\x98x\xcbu\x1a<\xc7Jf\x22\xec\x05\x92K\xbdE\
\xc5>\xdds=\xef\x90\xac~\xbc\xee\xfe\xa9\xf5f\xdd\
]Q\x1f\xf8\xee\xa5\x88\xe4\xe2\xac$+7\xee\x8eF\
\xa3\x024\x92\x02{\x0b\x11c\xc2\x90r\x16T\xd6g\
\xc0\xe2)\x8e\x89\x90\xea\x85\xf2\x05\xecJ\x8b\xc6N\xf5\
\x06^\xfc\xab\x89\xf9\x1d\xdf\xaem\xfb\x9fl\xbb-|\
}yx\x92_6\xe0\x8e\xa7\xf2;\xa86\xf4\x93\xa1\
\x11\xdd\xa3\x9b\xd0C\x96R\x9f\xebl\xb6\x10\xf4\x03#\
\x81\xdb\x10\xfeW\xc16\x1b\x93\xdbb@\xcem\xd9\x88\
l\xb2&7\xc6\x80\x9f\x1b\xb3\x19\xdflNn\x8d\x81\
\xc2]I(J\xf1.\x83\x0c\x1e\xa5z\x06U\xad\xc8\
\x12u\xc6B\x9c\x8e\x89\xb3\xd6\x04ee\xa6\x92\xb3\xb0\
\xb1;\xfa\xf9\x86V\x8e\xa7\xd9\xe2|2O\xe0\xac\xf3\
\x00\x8aE\xf3\xb5\xbe[\xc6;\xf0\xbd\xa1\x1b\xb4\xca\xe9\
{n\xe0\x0c\xb6\xa5\xba\xd6\x99i#r\xed\x817h\
'r\xec\xbe\xdb\xaf\x12\xdd*\x19+H\xd3\xedM[\
N\x8enC/\xfb\x92\x87\xd5\xd6g?\xb4\xbf\xce \
\xc1\xea\xd4\x08\xa4xda\x0cx\x8f\x14\xb3\x1au\x0b\
X\x19\xfel\xca\x10\x8c\xe3\xca>$\xc5w\xb0\x98T\
\x86\xb0\xcbe\x94I\x8fC\x84\xcb\xeb\x8e\xf3\xe7\x0d*\
\x8f\xa7~\x87VSh\xa8\xc8\xb9f\x82\x15g\xf5y\
\xa8#\xdbs\xfb'z\x84\x96?\xfb\xef\xab\xc1\x97\xba\
\xa1^\xea2F\x8d\xb7\xd4\x85\x8c\xba\x8dY\x990\xe4\
\xf0!I\xb1\xebO\x8f\x15\xa3\x94\x1cY\xd2\xf0(\xf9\
\x07\xf0\xea4^24L\xcbEe\x17\xbdO|\xae\
\xc1\x1b\xadv\xec\xee\xdf$\xcb\xcco\x98\xa9\xb1\xa2\xf3\
\x8a7W,\xc5\xed\xb3\xd1\xb0\xed\x03[k\xb8V\x96\
\xdb\x19\xe9\xd5FiE\xdf{\xe4\xfe\xffB\xaadl\
\x15W\xadb\x10^\xb1\xa7__b1&y\x9b5\
*1\xbe\x0d\xd5i\xee\x81C\x80\x0f\x0cf\x86\xf2\x14\
R\xb4\xcaGEE\x9aP\x05\x02\xd7\xb1\x8b\xc3\x15\x13\
p\x11\xb25d3N\xe3\xfbbB\xc3\x90\x89\xb8\x5c\
y\xb3~\x01\xbaga\x94\
"

qt_resource_name = b"\
\x00\x0a\
\x0a\xf6\x0f\xa3\
\x00s\
\x00t\x00a\x00t\x00i\x00c\x00_\x00c\x00s\x00s\
\x00\x0e\
\x03\xc2\xd5\xc3\
\x00s\
\x00t\x00y\x00l\x00e\x00s\x00h\x00e\x00e\x00t\x00.\x00q\x00s\x00s\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x1a\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01\x92\x09g[\xe6\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()



from PySide6.QtCore import Qt
from PySide6.QtGui import (
    QPixmap,
    QImage,
)
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QSpacerItem,
    QSizePolicy,
    QHBoxLayout,
    QLabel,
)

from rare.utils.qt_requests import QtRequests
from rare.widgets.image_widget import ImageWidget
from rare.widgets.loading_widget import LoadingWidget


class IconWidget(object):
    def __init__(self):
        self.mini_widget: QWidget = None
        self.title_label: QLabel = None
        self.developer_label: QLabel = None
        self.price_label: QLabel = None
        self.discount_label: QLabel = None

    def setupUi(self, widget: QWidget):
        # on-hover popup
        self.mini_widget = QWidget(parent=widget)
        self.mini_widget.setObjectName(f"{type(self).__name__}MiniWidget")
        self.mini_widget.setFixedHeight(int(widget.height() // 3))

        # game title
        self.title_label = QLabel(parent=self.mini_widget)
        self.title_label.setObjectName(f"{type(self).__name__}TitleLabel")
        self.title_label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.title_label.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.title_label.setAutoFillBackground(False)
        self.title_label.setWordWrap(True)

        # information below title
        self.developer_label = QLabel(parent=self.mini_widget)
        self.developer_label.setObjectName(f"{type(self).__name__}TooltipLabel")
        self.developer_label.setAlignment(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
        self.developer_label.setAutoFillBackground(False)

        self.price_label = QLabel(parent=self.mini_widget)
        self.price_label.setObjectName(f"{type(self).__name__}TooltipLabel")
        self.price_label.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self.price_label.setAutoFillBackground(False)

        self.discount_label = QLabel(parent=self.mini_widget)
        self.discount_label.setObjectName(f"{type(self).__name__}TooltipLabel")
        self.discount_label.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self.discount_label.setAutoFillBackground(False)

        # Create layouts
        # layout on top of the image, holds the status label, a spacer item and the mini widget
        image_layout = QVBoxLayout()
        image_layout.setContentsMargins(2, 2, 2, 2)

        # layout for the mini widget, holds the top row and the info label
        mini_layout = QVBoxLayout()
        mini_layout.setSpacing(0)

        # layout for the top row, holds the title and the launch button
        row_layout = QHBoxLayout()
        row_layout.setSpacing(6)
        row_layout.setAlignment(Qt.AlignmentFlag.AlignBottom)

        # Layout the widgets
        # (from inner to outer)
        row_layout.addWidget(self.developer_label, stretch=2)
        row_layout.addWidget(self.price_label)
        row_layout.addWidget(self.discount_label)
        mini_layout.addWidget(self.title_label)
        mini_layout.addLayout(row_layout)
        self.mini_widget.setLayout(mini_layout)

        image_layout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))
        image_layout.addWidget(self.mini_widget)
        widget.setLayout(image_layout)


class LoadingImageWidget(ImageWidget):
    def __init__(self, manager: QtRequests, parent=None):
        super(LoadingImageWidget, self).__init__(parent=parent)
        self.ui = IconWidget()
        self.spinner = LoadingWidget(parent=self)
        self.spinner.setVisible(False)
        self.manager = manager

    def fetchPixmap(self, url):
        self.setPixmap(QPixmap())
        self.spinner.setFixedSize(self._image_size.size)
        self.spinner.start()
        self.manager.get(url, self.__on_image_ready, params={
            "resize": 1,
            "w": self._image_size.base.size.width(),
            "h": self._image_size.base.size.height(),
        })

    def __on_image_ready(self, data):
        cover = QImage()
        cover.loadFromData(data)
        # cover = cover.scaled(self._image_size.size, Qt.KeepAspectRatio, Qt.SmoothTransformation)
        cover.setDevicePixelRatio(self._image_size.base.pixel_ratio)
        cover = cover.convertToFormat(QImage.Format.Format_ARGB32_Premultiplied)
        cover = QPixmap(cover)
        self.setPixmap(cover)
        self.spinner.stop()



import os
from dataclasses import dataclass
from logging import getLogger
from typing import Dict, Tuple, List, Optional

logger = getLogger("Wine")

lutris_runtime_paths = [
    os.path.expanduser("~/.local/share/lutris")
]

__lutris_runtime: str = None
__lutris_wine: str = None


def find_lutris() -> Tuple[str, str]:
    global __lutris_runtime, __lutris_wine
    for path in lutris_runtime_paths:
        runtime_path = os.path.join(path, "runtime")
        wine_path = os.path.join(path, "runners", "wine")
        if os.path.isdir(path) and os.path.isdir(runtime_path) and os.path.isdir(wine_path):
            __lutris_runtime, __lutris_wine = runtime_path, wine_path
            return runtime_path, wine_path


@dataclass
class WineRuntime:
    name: str
    path: str
    environ: Dict


@dataclass
class WineRunner:
    name: str
    path: str
    environ: Dict
    runtime: Optional[WineRuntime] = None


def find_lutris_wines(runtime_path: str = None, wine_path: str = None) -> List[WineRunner]:
    runners = []
    if not runtime_path and not wine_path:
        return runners


def __get_lib_path(executable: str, basename: str = "") -> str:
    path = os.path.dirname(os.path.dirname(executable))
    lib32 = os.path.realpath(os.path.join(path, "lib32", basename))
    lib64 = os.path.realpath(os.path.join(path, "lib64", basename))
    lib = os.path.realpath(os.path.join(path, "lib", basename))
    if lib32 == lib or not os.path.exists(lib32):
        ldpath = ":".join([lib64, lib])
    elif lib64 == lib or not os.path.exists(lib64):
        ldpath = ":".join([lib, lib32])
    else:
        ldpath = lib if os.path.exists(lib) else lib64
    return ldpath


def get_wine_environment(executable: str = None, prefix: str = None) -> Dict:
    # If the tool is unset, return all affected env variable names
    # IMPORTANT: keep this in sync with the code below
    environ = {"WINEPREFIX": prefix if prefix is not None else ""}
    if executable is None:
        environ["WINEDLLPATH"] = ""
        environ["LD_LIBRARY_PATH"] = ""
    else:
        winedllpath = __get_lib_path(executable, "wine")
        environ["WINEDLLPATH"] = winedllpath
        librarypath = __get_lib_path(executable, "")
        environ["LD_LIBRARY_PATH"] = librarypath
    return environ


if __name__ == "__main__":
    from pprint import pprint

    pprint(get_wine_environment(
        "/opt/wine-ge-custom/bin/wine", None))
    pprint(get_wine_environment(
        "/usr/bin/wine", None))
    pprint(get_wine_environment(
        "/usr/share/steam/compatitiblitytools.d/dist/bin/wine", None))
    pprint(get_wine_environment(
        os.path.expanduser("~/.local/share/Steam/compatibilitytools.d/GE-Proton8-14/files/bin/wine"), None))
    pprint(get_wine_environment(
        os.path.expanduser("~/.local/share/lutris/runners/wine/lutris-GE-Proton8-14-x86_64/bin/wine"), None))



from PySide6.QtCore import QEvent, QPoint
from PySide6.QtGui import QIcon, QResizeEvent
from PySide6.QtWidgets import QTabBar, QPushButton, QSizePolicy


class MainTabBar(QTabBar):
    def __init__(self, parent=None):
        super(MainTabBar, self).__init__(parent=parent)
        self.setObjectName(type(self).__name__)
        font = self.font()
        font.setPointSize(font.pointSize() + 1)
        font.setBold(True)
        self.setFont(font)

        self.expanded = -1
        self.button: TabButtonWidget = None

    def tabSizeHint(self, index):
        size = super(MainTabBar, self).tabSizeHint(index)
        if index == self.expanded:
            offset = self.width()
            for index in range(self.count()):
                offset -= super(MainTabBar, self).tabSizeHint(index).width()
            size.setWidth(max(size.width(), size.width() + offset))
        return size

    def __align_button(self):
        self.button.move(QPoint(
            self.tabRect(self.expanded).right() - self.button.width() - self.style().PixelMetric.PM_DefaultFrameWidth,
            self.tabRect(self.expanded).bottom() - self.button.height()
        ))

    def resizeEvent(self, e: QResizeEvent):
        super().resizeEvent(e)
        self.__align_button()

    def event(self, e):
        if e.type() == QEvent.Type.StyleChange:
            self.button.setFixedHeight(self.button.minimumSizeHint().height())
            self.button.setFixedWidth(self.button.minimumSizeHint().width())
            self.__align_button()
        return super().event(e)

    def setButton(self, widget: QPushButton):
        widget.setParent(self)
        widget.setFixedHeight(widget.minimumSizeHint().height())
        widget.setFixedWidth(widget.minimumSizeHint().width())
        self.button = widget


class TabButtonWidget(QPushButton):
    def __init__(self, icon: QIcon, text: str = "", tooltip: str = "", parent=None):
        super(TabButtonWidget, self).__init__(parent=parent)
        self.setObjectName(type(self).__name__)
        self.setIcon(icon)
        self.setText(text)
        self.setToolTip(tooltip)
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)



import os
from enum import IntEnum
from logging import getLogger
from typing import Union, Type, Dict, Tuple, Iterable

import qtawesome
from PySide6.QtCore import (
    QObject,
    QSettings,
    QFile,
    Qt,
    QLocale,
    QDirIterator,
)
from PySide6.QtGui import QPalette, QColor, QFontMetrics
from PySide6.QtWidgets import QApplication, QStyleFactory, QLabel
from shiboken6.Shiboken import Object as ShibokenObject

from rare.utils.paths import resources_path

logger = getLogger("Utils")


class ExitCodes(IntEnum):
    EXIT = 0
    LOGOUT = -133742


color_role_map: Dict[int, str] = {
    0: "WindowText",
    1: "Button",
    2: "Light",
    3: "Midlight",
    4: "Dark",
    5: "Mid",
    6: "Text",
    7: "BrightText",
    8: "ButtonText",
    9: "Base",
    10: "Window",
    11: "Shadow",
    12: "Highlight",
    13: "HighlightedText",
    14: "Link",
    15: "LinkVisited",
    16: "AlternateBase",
    # 17: "NoRole",
    18: "ToolTipBase",
    19: "ToolTipText",
    20: "PlaceholderText",
    # 21: "NColorRoles",
}

color_group_map: Dict[int, str] = {
    0: "Active",
    1: "Disabled",
    2: "Inactive",
}


def load_color_scheme(path: str) -> QPalette:
    palette = QPalette()
    scheme = QSettings(path, QSettings.Format.IniFormat)
    try:
        scheme.beginGroup("ColorScheme")
        for g in color_group_map:
            scheme.beginGroup(color_group_map[g])
            group = QPalette.ColorGroup(g)
            for r in color_role_map:
                role = QPalette.ColorRole(r)
                color = scheme.value(color_role_map[r], None)
                if color is not None:
                    palette.setColor(group, role, QColor(color))
                else:
                    palette.setColor(group, role, palette.color(QPalette.ColorGroup.Active, role))
            scheme.endGroup()
        scheme.endGroup()
    except:
        palette = None
    return palette


def get_static_style() -> str:
    file = QFile(":/static_css/stylesheet.qss")
    file.open(QFile.OpenModeFlag.ReadOnly)
    static = file.readAll().data().decode("utf-8")
    file.close()
    return static


def set_color_pallete(color_scheme: str) -> None:
    static = get_static_style()
    qApp: QApplication = QApplication.instance()

    if not color_scheme:
        qApp.setStyle(QStyleFactory.create(qApp.property("rareDefaultQtStyle")))
        qApp.setPalette(qApp.style().standardPalette())
        qApp.setStyleSheet(static)
        return

    qApp.setStyle(QStyleFactory.create("Fusion"))
    custom_palette = load_color_scheme(f":/schemes/{color_scheme}")
    if custom_palette is not None:
        qApp.setPalette(custom_palette)
        qApp.setStyleSheet(static)
        icon_color_normal = qApp.palette().color(QPalette.ColorRole.WindowText).name()
        icon_color_disabled = qApp.palette().color(QPalette.ColorRole.WindowText).name()
        qtawesome.set_defaults(color=icon_color_normal, color_disabled=icon_color_disabled)


def get_color_schemes() -> Iterable[str]:
    it = QDirIterator(":/schemes/")
    while it.hasNext():
        it.next()
        yield it.fileName()


def set_style_sheet(style_sheet: str) -> None:
    static = get_static_style()
    qApp: QApplication = QApplication.instance()

    if not style_sheet:
        qApp.setStyle(QStyleFactory.create(qApp.property("rareDefaultQtStyle")))
        qApp.setStyleSheet(static)
        return

    qApp.setStyle(QStyleFactory.create("Fusion"))
    file = QFile(f":/stylesheets/{style_sheet}/stylesheet.qss")
    file.open(QFile.OpenModeFlag.ReadOnly)
    stylesheet = file.readAll().data().decode("utf-8")
    file.close()
    qApp.setStyleSheet(stylesheet + static)

    icon_color_normal = qApp.palette().color(QPalette.ColorRole.Text).name()
    icon_color_disabled = qApp.palette().color(QPalette.ColorRole.Text).name()
    qtawesome.set_defaults(color="#eee", color_disabled="#eee")


def get_style_sheets() -> Iterable[str]:
    it = QDirIterator(":/stylesheets/")
    while it.hasNext():
        it.next()
        yield it.fileName()


def get_translations() -> Tuple[Tuple[str, str], ...]:
    langs = []
    for i in os.listdir(os.path.join(resources_path, "languages")):
        if i.endswith(".qm") and i.startswith("rare_"):
            locale = QLocale(i.removesuffix(".qm").removeprefix("rare_"))
            langs.append((locale.name(), f"{locale.nativeLanguageName()} ({locale.nativeCountryName()})"))
    return tuple(langs)


def path_size(path: Union[str, os.PathLike]) -> int:
    return sum(
        os.stat(os.path.join(dp, f)).st_size
        for dp, dn, filenames in os.walk(path)
        for f in filenames
        if os.path.isfile(os.path.join(dp, f))
    )


def format_size(b: Union[int, float]) -> str:
    for s in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei"):
        if b < 1024:
            return f"{b:.2f} {s}B"
        b /= 1024


def qta_icon(icn_str: str, fallback: str = None, **kwargs):
    try:
        return qtawesome.icon(icn_str, **kwargs)
    except Exception as e:
        if not fallback:
            logger.warning(f"{e} {icn_str}")
    if fallback:
        try:
            return qtawesome.icon(fallback, **kwargs)
        except Exception as e:
            logger.error(str(e))
    if kwargs.get("color"):
        kwargs["color"] = "red"
    return qtawesome.icon("ei.error", **kwargs)


def widget_object_name(widget: Union[QObject, ShibokenObject, Type], suffix: str) -> str:
    suffix = f"_{suffix}" if suffix else ""
    if isinstance(widget, QObject):
        return f"{type(widget).__name__}{suffix}"
    elif isinstance(widget, ShibokenObject) or isinstance(widget, type):
        return f"{widget.__name__}{suffix}"
    else:
        raise RuntimeError(f"Argument {widget} not a QObject or type of QObject")


def elide_text(label: QLabel, text: str) -> str:
    metrics = QFontMetrics(label.font())
    return metrics.elidedText(text, Qt.TextElideMode.ElideRight, label.sizeHint().width())


def style_hyperlink(link: str, title: str) -> str:
    return "<a href='{}' style='color: #2980b9; text-decoration:none'>{}</a>".format(link, title)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'login_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QHBoxLayout, QLabel,
    QPushButton, QSizePolicy, QSpacerItem, QVBoxLayout,
    QWidget)

class Ui_LoginDialog(object):
    def setupUi(self, LoginDialog):
        if not LoginDialog.objectName():
            LoginDialog.setObjectName(u"LoginDialog")
        LoginDialog.resize(241, 128)
        self.main_layout = QVBoxLayout(LoginDialog)
        self.main_layout.setObjectName(u"main_layout")
        self.login_vspacer_top = QSpacerItem(0, 17, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)

        self.main_layout.addItem(self.login_vspacer_top)

        self.welcome_label = QLabel(LoginDialog)
        self.welcome_label.setObjectName(u"welcome_label")

        self.main_layout.addWidget(self.welcome_label)

        self.login_vspacer_bottom = QSpacerItem(0, 17, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)

        self.main_layout.addItem(self.login_vspacer_bottom)

        self.login_stack_layout = QVBoxLayout()
        self.login_stack_layout.setObjectName(u"login_stack_layout")

        self.main_layout.addLayout(self.login_stack_layout)

        self.button_layout = QHBoxLayout()
        self.button_layout.setObjectName(u"button_layout")
        self.exit_button = QPushButton(LoginDialog)
        self.exit_button.setObjectName(u"exit_button")

        self.button_layout.addWidget(self.exit_button)

        self.button_hspacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.button_layout.addItem(self.button_hspacer)

        self.back_button = QPushButton(LoginDialog)
        self.back_button.setObjectName(u"back_button")

        self.button_layout.addWidget(self.back_button)

        self.next_button = QPushButton(LoginDialog)
        self.next_button.setObjectName(u"next_button")

        self.button_layout.addWidget(self.next_button)


        self.main_layout.addLayout(self.button_layout)


        self.retranslateUi(LoginDialog)
    # setupUi

    def retranslateUi(self, LoginDialog):
        LoginDialog.setWindowTitle(QCoreApplication.translate("LoginDialog", u"Login", None))
        self.welcome_label.setText(QCoreApplication.translate("LoginDialog", u"<h1>Welcome to Rare</h1>", None))
        self.exit_button.setText(QCoreApplication.translate("LoginDialog", u"Exit", None))
        self.back_button.setText(QCoreApplication.translate("LoginDialog", u"Back", None))
        self.next_button.setText(QCoreApplication.translate("LoginDialog", u"Next", None))
    # retranslateUi




from setuptools import setup

from setuptools_scm import ScmVersion


def mkversion(ver: ScmVersion) -> str:
    return f"{ver.tag}.{ver.distance}"


setup(use_scm_version={"version_scheme": mkversion})



#!/usr/bin/env python3
import os

for f in os.listdir(os.path.join(os.path.dirname(__file__), "../rare/resources/languages/")):
    if f.endswith(".ts") and f != "translation_source.ts":
        os.system(f"pyside6-lrelease {os.path.join(os.path.dirname(__file__), '../rare/resources/languages/', f)}")



from enum import IntEnum


class LibraryView(IntEnum):
    COVER = 1
    VLIST = 2


class LibraryFilter(IntEnum):
    ALL = 1
    INSTALLED = 2
    OFFLINE = 3
    HIDDEN = 4
    WIN32 = 5
    MAC = 6
    INSTALLABLE = 7
    INCLUDE_UE = 8


class LibraryOrder(IntEnum):
    TITLE = 1
    RECENT = 2
    NEWEST = 3
    OLDEST = 4



import platform as pf
from typing import Type

from PySide6.QtCore import QSettings, Qt
from PySide6.QtGui import QHideEvent
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout
)

from rare.shared import LegendaryCoreSingleton
from rare.utils import config_helper as config
from rare.widgets.side_tab import SideTabContents
from .env_vars import EnvVars
from .launch import LaunchSettingsType
from .overlay import DxvkSettings

if pf.system() != "Windows":
    from .wine import WineSettings

if pf.system() in {"Linux", "FreeBSD"}:
    from .proton import ProtonSettings
    from .overlay import MangoHudSettings


class GameSettingsBase(QWidget, SideTabContents):

    def __init__(
        self,
        launch_widget: Type[LaunchSettingsType],
        dxvk_widget: Type[DxvkSettings],
        envvar_widget: Type[EnvVars],
        wine_widget: Type['WineSettings'] = None,
        proton_widget: Type['ProtonSettings'] = None,
        mangohud_widget: Type['MangoHudSettings'] = None,
        parent=None
    ):
        super(GameSettingsBase, self).__init__(parent=parent)

        self.core = LegendaryCoreSingleton()
        self.settings = QSettings(self)
        self.app_name: str = "default"

        self.launch = launch_widget(self)
        self.env_vars = envvar_widget(self)

        if pf.system() != "Windows":
            self.wine = wine_widget(self)
            self.wine.environ_changed.connect(self.env_vars.reset_model)

        if pf.system() in {"Linux", "FreeBSD"}:
            self.proton_tool = proton_widget(self)
            self.proton_tool.environ_changed.connect(self.env_vars.reset_model)
            self.proton_tool.tool_enabled.connect(self.wine.tool_enabled)
            self.proton_tool.tool_enabled.connect(self.launch.tool_enabled)

        self.dxvk = dxvk_widget(self)
        self.dxvk.environ_changed.connect(self.env_vars.reset_model)

        if pf.system() in {"Linux", "FreeBSD"}:
            self.mangohud = mangohud_widget(self)
            self.mangohud.environ_changed.connect(self.env_vars.reset_model)

        self.main_layout = QVBoxLayout(self)
        self.main_layout.addWidget(self.launch)
        if pf.system() != "Windows":
            self.main_layout.addWidget(self.wine)
        if pf.system() in {"Linux", "FreeBSD"}:
            self.main_layout.addWidget(self.proton_tool)
        self.main_layout.addWidget(self.dxvk)
        if pf.system() in {"Linux", "FreeBSD"}:
            self.main_layout.addWidget(self.mangohud)
        self.main_layout.addWidget(self.env_vars)

        self.main_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

    def hideEvent(self, a0: QHideEvent):
        if a0.spontaneous():
            return super().hideEvent(a0)
        config.save_config()
        return super().hideEvent(a0)



import json
import platform
import shlex
import subprocess
import time
import traceback
from argparse import Namespace
from logging import getLogger
from signal import signal, SIGINT, SIGTERM, strsignal
from typing import Optional

# from PySide6.import sip
from PySide6.QtCore import (
    QObject,
    QProcess,
    Signal,
    QUrl,
    QRunnable,
    QThreadPool,
    QSettings,
    Qt,
    Slot,
    QTimer,
    QMetaMethod,
)
from PySide6.QtGui import QDesktopServices
from PySide6.QtNetwork import QLocalServer, QLocalSocket
from PySide6.QtWidgets import QApplication
from legendary.models.game import SaveGameStatus

from rare.lgndr.core import LegendaryCore
from rare.models.base_game import RareGameSlim
from rare.models.launcher import ErrorModel, Actions, FinishedModel, BaseModel, StateChangedModel
from rare.models.options import options
from rare.widgets.rare_app import RareApp, RareAppException
from .cloud_sync_dialog import CloudSyncDialog, CloudSyncDialogResult
from .console_dialog import ConsoleDialog
from .lgd_helper import get_launch_args, InitArgs, get_configured_process, LaunchArgs, GameArgsError

DETACHED_APP_NAMES = {
    "0a2d9f6403244d12969e11da6713137b",  # Fall Guys
    "Fortnite",
    "afdb5a85efcc45d8ae8e406e2121d81c",  # Fortnite Battle Royale
    "09e442f830a341f698b4da42abd98c9b",  # Fortnite Festival
    "d8f7763e07d74c209d760a679f9ed6ac",  # Lego Fortnite
    "Fortnite_Studio",                   # Unreal Editor for Fortnite
    "dcfccf8d965a4f2281dddf9fead042de",  # Homeworld Remastered Collection (issue#376)
}


class PreLaunch(QRunnable):
    class Signals(QObject):
        ready_to_launch = Signal(LaunchArgs)
        pre_launch_command_started = Signal()
        pre_launch_command_finished = Signal(int)  # exit_code
        error_occurred = Signal(str)

    def __init__(self, args: InitArgs, rgame: RareGameSlim, sync_action=None):
        super(PreLaunch, self).__init__()
        self.signals = self.Signals()
        self.logger = getLogger(type(self).__name__)
        self.args = args
        self.rgame = rgame
        self.sync_action = sync_action

    def run(self) -> None:
        self.logger.info(f"Sync action: {self.sync_action}")
        if self.sync_action == CloudSyncDialogResult.UPLOAD:
            self.rgame.upload_saves(False)
        elif self.sync_action == CloudSyncDialogResult.DOWNLOAD:
            self.rgame.download_saves(False)
        else:
            self.logger.info("No sync action")

        if args := self.prepare_launch(self.args):
            self.signals.ready_to_launch.emit(args)
        else:
            return

    def prepare_launch(self, args: InitArgs) -> Optional[LaunchArgs]:
        try:
            launch_args = get_launch_args(self.rgame, args)
        except Exception as e:
            self.signals.error_occurred.emit(str(e))
            return None
        if not launch_args:
            return None

        if launch_args.pre_launch_command:
            proc = get_configured_process()
            proc.setProcessEnvironment(launch_args.environment)
            self.signals.pre_launch_command_started.emit()
            prelaunch = shlex.split(launch_args.pre_launch_command)
            command = prelaunch.pop(0) if len(prelaunch) else ""
            arguments = prelaunch if len(prelaunch) else []
            self.logger.info("Running pre-launch command %s, %s", command, shlex.join(arguments))
            if launch_args.pre_launch_wait:
                proc.start(command, arguments)
                self.logger.info("Waiting for pre-launch command to finish")
                proc.waitForFinished(-1)
            else:
                proc.startDetached(command, arguments)
        return launch_args


class SyncCheckWorker(QRunnable):
    class Signals(QObject):
        sync_state_ready = Signal()
        error_occurred = Signal(str)

    def __init__(self, core: LegendaryCore, rgame: RareGameSlim):
        super().__init__()
        self.signals = self.Signals()
        self.core = core
        self.rgame = rgame

    def run(self) -> None:
        try:
            self.rgame.update_saves()
        except Exception as e:
            self.signals.error_occurred.emit(str(e))
            return
        self.signals.sync_state_ready.emit()


class RareLauncherException(RareAppException):
    def __init__(self, app: "RareLauncher", args: Namespace, parent=None):
        super(RareLauncherException, self).__init__(parent=parent)
        self.__app = app
        self.__args = args

    def _handler(self, exc_type, exc_value, exc_tb) -> bool:
        try:
            self.__app.send_message(ErrorModel(
                app_name=self.__args.app_name,
                action=Actions.error,
                error_string="".join(traceback.format_exception(exc_type, exc_value, exc_tb))
            ))
        except RuntimeError:
            pass
        return False


class RareLauncher(RareApp):
    exit_app = Signal()

    def __init__(self, args: InitArgs):
        super(RareLauncher, self).__init__(args, f"{type(self).__name__}_{args.app_name}_{{0}}.log")
        self.socket: Optional[QLocalSocket] = None
        self.console: Optional[ConsoleDialog] = None
        self.game_process: QProcess = QProcess(self)
        self.server: QLocalServer = QLocalServer(self)

        self._hook.deleteLater()
        self._hook = RareLauncherException(self, args, self)

        self.success: bool = False
        self.no_sync_on_exit = False
        self.args = args
        self.core = LegendaryCore()
        game = self.core.get_game(args.app_name)
        if not game:
            self.logger.error(f"Game {args.app_name} not found. Exiting")
            return
        self.rgame = RareGameSlim(self.core, game, self)

        language = self.settings.value(*options.language)
        self.load_translator(language)

        # if QSettings(self).value(*options.log_games) or (
        #     game.app_name in DETACHED_APP_NAMES and platform.system() == "Windows"
        # ):
        if QSettings(self).value(*options.log_games):
            self.console = ConsoleDialog(game.app_title)
            self.console.show()

        self.sync_dialog: Optional[CloudSyncDialog] = None

        self.game_process.finished.connect(self.__process_finished)
        self.game_process.errorOccurred.connect(self.__process_errored)
        if self.console:
            self.game_process.readyReadStandardOutput.connect(self.__proc_log_stdout)
            self.game_process.readyReadStandardError.connect(self.__proc_log_stderr)
            self.console.term.connect(self.__proc_term)
            self.console.kill.connect(self.__proc_kill)

        ret = self.server.listen(f"rare_{args.app_name}")
        if not ret:
            self.logger.error(self.server.errorString())
            self.logger.info("Server is running")
            self.server.close()
            return
        self.server.newConnection.connect(self.new_server_connection)

        self.success = True
        self.start_time = time.time()

        # This launches the application after it has been instantiated.
        # The timer's signal will be serviced once we call `exec()` on the application
        QTimer.singleShot(0, self.start)

    @Slot()
    def __proc_log_stdout(self):
        self.console.log_stdout(
            self.game_process.readAllStandardOutput().data().decode("utf-8", "ignore")
        )

    @Slot()
    def __proc_log_stderr(self):
        self.console.log_stderr(
            self.game_process.readAllStandardError().data().decode("utf-8", "ignore")
        )

    @Slot()
    def __proc_term(self):
        self.game_process.terminate()

    @Slot()
    def __proc_kill(self):
        self.game_process.kill()

    def new_server_connection(self):
        if self.socket is not None:
            try:
                self.socket.disconnectFromServer()
            except RuntimeError:
                pass
        self.logger.info("New connection")
        self.socket = self.server.nextPendingConnection()
        self.socket.disconnected.connect(self.socket_disconnected)
        self.socket.flush()

    def socket_disconnected(self):
        self.logger.info("Server disconnected")
        self.socket = None

    def send_message(self, message: BaseModel):
        if self.socket:
            self.socket.write(json.dumps(vars(message)).encode("utf-8"))
            self.socket.flush()
        else:
            self.logger.error("Can't send message")

    def check_saves_finished(self, exit_code: int):
        self.rgame.signals.widget.update.connect(lambda: self.on_exit(exit_code))

        state, (dt_local, dt_remote) = self.rgame.save_game_state

        if state == SaveGameStatus.LOCAL_NEWER and not self.no_sync_on_exit:
            action = CloudSyncDialogResult.UPLOAD
            self.__check_saves_finished(exit_code, action)
        else:
            self.sync_dialog = CloudSyncDialog(self.rgame.igame, dt_local, dt_remote)
            self.sync_dialog.result_ready.connect(lambda a: self.__check_saves_finished(exit_code, a))
            self.sync_dialog.open()

    @Slot(int, int)
    @Slot(int, CloudSyncDialogResult)
    def __check_saves_finished(self, exit_code, action):
        if self.sync_dialog is not None:
            self.sync_dialog.deleteLater()
            self.sync_dialog = None
        action = CloudSyncDialogResult(action)
        if action == CloudSyncDialogResult.UPLOAD:
            if self.console:
                self.console.log("Uploading saves...")
            self.rgame.upload_saves()
        elif action == CloudSyncDialogResult.DOWNLOAD:
            if self.console:
                self.console.log("Downloading saves...")
            self.rgame.download_saves()
        else:
            self.on_exit(exit_code)

    @Slot(int, QProcess.ExitStatus)
    def __process_finished(self, exit_code: int, exit_status: QProcess.ExitStatus):
        self.logger.info("Game finished")

        if self.rgame.auto_sync_saves:
            self.check_saves_finished(exit_code)
        else:
            self.on_exit(exit_code)

    @Slot(QProcess.ProcessError)
    def __process_errored(self, error: QProcess.ProcessError):
        self.error_occurred(self.game_process.errorString())

    def on_exit(self, exit_code: int):
        if self.console:
            self.console.on_process_exit(self.core.get_game(self.rgame.app_name).app_title, exit_code)

        self.send_message(FinishedModel(
            action=Actions.finished,
            app_name=self.rgame.app_name,
            exit_code=exit_code,
            playtime=int(time.time() - self.start_time),
        ))
        self.stop()

    @Slot(object)
    def launch_game(self, args: LaunchArgs):
        # should never happen
        if not args:
            self.stop()
            return
        if self.console:
            self.console.set_env(args.environment)
        self.start_time = time.time()

        if self.args.dry_run:
            self.logger.info("Dry run %s (%s)", self.rgame.app_title, self.rgame.app_name)
            self.logger.info("%s %s", args.executable, " ".join(args.arguments))
            if self.console:
                self.console.log(f"Dry run {self.rgame.app_title} ({self.rgame.app_name})")
                self.console.log(f"{shlex.join([args.executable] + args.arguments)}")
                self.console.accept_close = True
            print(shlex.join([args.executable] + args.arguments))
            self.stop()
            return

        if args.is_origin_game:
            QDesktopServices.openUrl(QUrl(args.executable))
            self.stop()  # stop because it is not a subprocess
            return

        self.logger.debug("Launch command %s, %s", args.executable, " ".join(args.arguments))
        self.logger.debug("Working directory %s", args.working_directory)

        if self.rgame.app_name in DETACHED_APP_NAMES and platform.system() == "Windows":
            if self.console:
                self.console.log("Launching as a detached process")
            subprocess.Popen([args.executable] + args.arguments, cwd=args.working_directory,
                             env={i: args.environment.value(i) for i in args.environment.keys()})
            self.stop()  # stop because we do not attach to the output
            return

        if args.working_directory:
            self.game_process.setWorkingDirectory(args.working_directory)
        self.game_process.setProcessEnvironment(args.environment)
        # send start message after process started
        self.game_process.started.connect(lambda: self.send_message(StateChangedModel(
            action=Actions.state_update, app_name=self.rgame.app_name,
            new_state=StateChangedModel.States.started
        )))
        # self.logger.debug("Executing prelaunch command %s, %s", args.executable, args.arguments)
        self.game_process.start(args.executable, args.arguments)

    def error_occurred(self, error_str: str):
        self.logger.warning(error_str)
        if self.console:
            self.console.on_process_exit(self.core.get_game(self.rgame.app_name).app_title, error_str)
        self.send_message(ErrorModel(
            error_string=error_str, app_name=self.rgame.app_name,
            action=Actions.error
        ))
        self.stop()

    def start_prepare(self, sync_action=None):
        worker = PreLaunch(self.args, self.rgame, sync_action)
        worker.signals.ready_to_launch.connect(self.launch_game)
        worker.signals.error_occurred.connect(self.error_occurred)
        # worker.signals.started_pre_launch_command(None)

        QThreadPool.globalInstance().start(worker)

    def sync_ready(self):
        if self.rgame.is_save_up_to_date:
            if self.console:
                self.console.log("Sync worker ready. Sync not required")
            self.start_prepare()
            return

        _, (dt_local, dt_remote) = self.rgame.save_game_state
        self.sync_dialog = CloudSyncDialog(self.rgame.igame, dt_local, dt_remote)
        self.sync_dialog.result_ready.connect(self.__sync_ready)
        self.sync_dialog.open()

    @Slot(int)
    @Slot(CloudSyncDialogResult)
    def __sync_ready(self, action: CloudSyncDialogResult):
        if self.sync_dialog is not None:
            self.sync_dialog.deleteLater()
            self.sync_dialog = None
        action = CloudSyncDialogResult(action)
        if action == CloudSyncDialogResult.CANCEL:
            self.no_sync_on_exit = True
        if self.console:
            if action == CloudSyncDialogResult.DOWNLOAD:
                self.console.log("Downloading saves...")
            elif action == CloudSyncDialogResult.UPLOAD:
                self.console.log("Uploading saves...")
        self.start_prepare(action)

    def start(self):
        if not self.args.offline:
            try:
                if not self.core.login():
                    raise ValueError("You are not logged in")
            except ValueError:
                # automatically launch offline if available
                self.logger.error("Not logged in. Trying to launch the game in offline mode")
                self.args.offline = True

        if not self.args.offline and self.rgame.auto_sync_saves:
            self.logger.info("Start sync worker")
            worker = SyncCheckWorker(self.core, self.rgame)
            worker.signals.error_occurred.connect(self.error_occurred)
            worker.signals.sync_state_ready.connect(self.sync_ready)
            QThreadPool.globalInstance().start(worker)
        else:
            self.start_prepare()

    def stop(self):
        try:
            if self.console:
                self.game_process.readyReadStandardOutput.disconnect()
                self.game_process.readyReadStandardError.disconnect()
            if self.game_process.isSignalConnected(QMetaMethod.fromSignal(self.game_process.finished)):
                self.game_process.finished.disconnect()
            if self.game_process.isSignalConnected(QMetaMethod.fromSignal(self.game_process.errorOccurred)):
                self.game_process.errorOccurred.disconnect()
        except (TypeError, RuntimeError) as e:
            self.logger.error("Failed to disconnect process signals: %s", e)

        if self.game_process.state() != QProcess.ProcessState.NotRunning:
            self.game_process.kill()
        exit_code = self.game_process.exitCode()
        self.game_process.deleteLater()

        self.logger.info("Stopping server %s", self.server.socketDescriptor())
        try:
            self.server.close()
            self.server.deleteLater()
        except RuntimeError as e:
            self.logger.error("Error occured while stopping server: %s", e)

        self.processEvents()
        if not self.console:
            self.exit(exit_code)
        else:
            self.console.on_process_exit(self.rgame.app_title, exit_code)


def launch(args: Namespace) -> int:
    args = InitArgs.from_argparse(args)

    QApplication.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps, True)

    app = RareLauncher(args)
    app.setQuitOnLastWindowClosed(True)

    # This prevents ghost QLocalSockets, which block the name, which makes it unable to start
    # No handling for SIGKILL
    def sighandler(s, frame):
        app.logger.info("%s received. Stopping", strsignal(s))
        app.stop()
        app.exit(1)

    signal(SIGINT, sighandler)
    signal(SIGTERM, sighandler)

    if not app.success:
        app.stop()
        app.exit(1)
        return 1

    try:
        exit_code = app.exec()
    except Exception as e:
        app.logger.error("Unhandled error %s", e)
        exit_code = 1
    finally:
        pass
        # if not sip.isdeleted(app.server):
        #     app.server.close()
    return exit_code



import os
import platform
import shutil
from argparse import Namespace
from dataclasses import dataclass, field
from logging import getLogger
from typing import List

from PySide6.QtCore import QProcess, QProcessEnvironment
from legendary.models.game import LaunchParameters

from rare.models.base_game import RareGameSlim

logger = getLogger("RareLauncherUtils")


class GameArgsError(Exception):
    pass


class InitArgs(Namespace):
    app_name: str
    dry_run: bool = False
    debug: bool = False
    offline: bool = False
    skip_update_check: bool = False
    wine_prefix: str = ""
    wine_bin: str = ""

    @classmethod
    def from_argparse(cls, args):
        return cls(
            app_name=args.app_name,
            debug=args.debug,
            offline=args.offline,
            skip_update_check=args.skip_update_check,
            wine_bin=args.wine_bin,
            wine_prefix=args.wine_pfx,
            dry_run=args.dry_run
        )


@dataclass
class LaunchArgs:
    executable: str = ""
    arguments: List[str] = field(default_factory=list)
    working_directory: str = ""
    environment: QProcessEnvironment = None
    pre_launch_command: str = ""
    pre_launch_wait: bool = False
    is_origin_game: bool = False  # only for windows to launch as url

    def __bool__(self):
        return bool(self.executable)


def get_origin_params(rgame: RareGameSlim, init_args: InitArgs, launch_args: LaunchArgs) -> LaunchArgs:
    core = rgame.core
    app_name = rgame.app_name

    origin_uri = core.get_origin_uri(app_name, init_args.offline)
    if platform.system() == "Windows":
        launch_args.executable = origin_uri
        launch_args.arguments = []
        # only set it here true, because on linux it is a launch command like every other game
        launch_args.is_origin_game = True
        return launch_args

    command = core.get_app_launch_command(app_name)
    if not os.path.exists(command[0]) and shutil.which(command[0]) is None:
        return launch_args
    command.append(origin_uri)

    env = core.get_app_environment(app_name)
    launch_args.environment = QProcessEnvironment.systemEnvironment()

    if os.environ.get("container") == "flatpak":
        flatpak_command = ["flatpak-spawn", "--host"]
        flatpak_command.extend(f"--env={name}={value}" for name, value in env.items())
        command = flatpak_command + command
    else:
        for name, value in env.items():
            launch_args.environment.insert(name, value)

    launch_args.executable = command[0]
    launch_args.arguments = command[1:]

    return launch_args


def get_game_params(rgame: RareGameSlim, args: InitArgs, launch_args: LaunchArgs) -> LaunchArgs:
    if not args.offline:  # skip for update
        if not args.skip_update_check and not rgame.core.is_noupdate_game(rgame.app_name):
            try:
                latest = rgame.core.get_asset(rgame.app_name, rgame.igame.platform, update=False)
            except ValueError:
                raise GameArgsError("Metadata doesn't exist")
            else:
                if latest.build_version != rgame.igame.version:
                    raise GameArgsError("Game is not up to date. Please update first")

    if (not rgame.igame or not rgame.igame.executable) and rgame.game is not None:
        # override installed game with base title
        if rgame.is_launchable_addon:
            app_name = rgame.game.metadata['mainGameItem']['releaseInfo'][0]['appId']
            rgame.igame = rgame.core.get_installed_game(app_name)

    try:
        params: LaunchParameters = rgame.core.get_launch_parameters(
            app_name=rgame.igame.app_name, offline=args.offline, addon_app_name=rgame.game.app_name
        )
    except TypeError:
        logger.warning("Using older get_launch_parameters due to legendary version")
        params: LaunchParameters = rgame.core.get_launch_parameters(
            app_name=rgame.igame.app_name, offline=args.offline
        )

    full_params = []
    launch_args.environment = QProcessEnvironment.systemEnvironment()

    if os.environ.get("container") == "flatpak":
        full_params.extend(["flatpak-spawn", "--host"])
        full_params.extend(
            f"--env={name}={value}"
            for name, value in params.environment.items()
        )
    else:
        for name, value in params.environment.items():
            launch_args.environment.insert(name, value)

    full_params.extend(params.launch_command)
    full_params.append(os.path.join(params.game_directory, params.game_executable))
    full_params.extend(params.game_parameters)
    full_params.extend(params.egl_parameters)
    full_params.extend(params.user_parameters)

    launch_args.executable = full_params[0]
    launch_args.arguments = full_params[1:]
    launch_args.working_directory = params.working_directory

    return launch_args


def get_launch_args(rgame: RareGameSlim, init_args: InitArgs = None) -> LaunchArgs:

    resp = LaunchArgs()

    if not rgame.game:
        raise GameArgsError(f"Could not find metadata for {rgame.app_title}")

    if rgame.is_origin:
        init_args.offline = False
    else:
        if not rgame.is_installed:
            raise GameArgsError("Game is not installed or has unsupported format")

        if rgame.is_dlc and not rgame.is_launchable_addon:
            raise GameArgsError("Game is a DLC")
        if not os.path.exists(rgame.install_path):
            raise GameArgsError("Game path does not exist")

    if rgame.is_origin:
        resp = get_origin_params(rgame, init_args, resp)
    else:
        resp = get_game_params(rgame, init_args, resp)

    pre_cmd, wait = rgame.core.get_pre_launch_command(init_args.app_name)
    resp.pre_launch_command, resp.pre_launch_wait = pre_cmd, wait
    return resp


def get_configured_process(env: dict = None):
    proc = QProcess()
    proc.setProcessChannelMode(QProcess.MergedChannels)
    proc.readyReadStandardOutput.connect(
        lambda: logger.info(
            str(proc.readAllStandardOutput().data(), "utf-8", "ignore")
        )
    )
    environment = QProcessEnvironment.systemEnvironment()
    if env:
        for e in env:
            environment.insert(e, env[e])
    proc.setProcessEnvironment(environment)

    return proc



import os
import shlex
import shutil
from typing import Tuple, Type, TypeVar

from PySide6.QtCore import Qt, Slot
from PySide6.QtGui import QShowEvent
from PySide6.QtWidgets import QCheckBox, QFileDialog, QFormLayout, QVBoxLayout, QGroupBox, QLineEdit

from rare.shared import LegendaryCoreSingleton
import rare.utils.config_helper as config
from rare.widgets.indicator_edit import PathEdit, IndicatorReasonsCommon
from .wrappers import WrapperSettings


class LaunchSettingsBase(QGroupBox):

    def __init__(
        self,
        wrapper_widget: Type[WrapperSettings],
        parent=None
    ):
        super(LaunchSettingsBase, self).__init__(parent=parent)
        self.setTitle(self.tr("Launch settings"))

        self.core = LegendaryCoreSingleton()
        self.app_name: str = "default"

        self.prelaunch_cmd = PathEdit(
            path="",
            placeholder=self.tr("Path to a script or program to run before the game"),
            file_mode=QFileDialog.FileMode.ExistingFile,
            edit_func=self.__prelaunch_cmd_edit_callback,
            save_func=self.__prelaunch_cmd_save_callback,
        )

        self.prelaunch_args = QLineEdit("")
        self.prelaunch_args.setPlaceholderText(self.tr("Arguments to the script or program to run before the game"))
        self.prelaunch_args.setToolTip(self.prelaunch_args.placeholderText())
        self.prelaunch_args.textChanged.connect(self.__prelaunch_changed)

        self.prelaunch_check = QCheckBox(self.tr("Wait for the pre-launch command to finish before launching the game"))
        font = self.font()
        font.setItalic(True)
        self.prelaunch_check.setFont(font)
        self.prelaunch_check.stateChanged.connect(self.__prelauch_check_changed)

        prelaunch_layout = QVBoxLayout()
        prelaunch_layout.addWidget(self.prelaunch_cmd)
        prelaunch_layout.addWidget(self.prelaunch_args)
        prelaunch_layout.addWidget(self.prelaunch_check)

        self.main_layout = QFormLayout(self)
        self.main_layout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.main_layout.setLabelAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self.main_layout.setFormAlignment(Qt.AlignmentFlag.AlignLeading | Qt.AlignmentFlag.AlignTop)

        self.wrappers_widget = wrapper_widget(self)

        self.main_layout.addRow(self.tr("Wrappers"), self.wrappers_widget)
        self.main_layout.addRow(self.tr("Pre-launch"), prelaunch_layout)

    def showEvent(self, a0: QShowEvent):
        if a0.spontaneous():
            return super().showEvent(a0)
        prelaunch = shlex.split(config.get_option(self.app_name, "pre_launch_command", fallback=""))
        command = prelaunch.pop(0) if len(prelaunch) else ""
        arguments = prelaunch if len(prelaunch) else []
        wait = config.get_boolean(self.app_name, "pre_launch_wait", fallback=False)

        self.prelaunch_cmd.setText(command)
        self.prelaunch_args.setText(shlex.join(arguments))
        self.prelaunch_check.setChecked(wait)
        self.prelaunch_check.setEnabled(bool(command))

        return super().showEvent(a0)

    @Slot()
    def tool_enabled(self):
        self.wrappers_widget.update_state()

    @staticmethod
    def __prelaunch_cmd_edit_callback(text: str) -> Tuple[bool, str, int]:
        if not text.strip():
            return True, text, IndicatorReasonsCommon.VALID
        if not os.path.isfile(text) and not shutil.which(text):
            return False, text, IndicatorReasonsCommon.FILE_NOT_EXISTS
        else:
            return True, text, IndicatorReasonsCommon.VALID

    def __prelaunch_cmd_save_callback(self, text):
        self.prelaunch_check.setEnabled(bool(text))
        self.__prelaunch_changed()

    def __prelauch_check_changed(self):
        config.set_boolean(self.app_name, "pre_launch_wait", self.prelaunch_check.isChecked())

    @Slot()
    def __prelaunch_changed(self):
        command = self.prelaunch_cmd.text().strip()
        if not command:
            config.save_option(self.app_name, "pre_launch_command", command)
            config.remove_option(self.app_name, "pre_launch_wait")
            return
        command = shlex.quote(command)
        arguments = self.prelaunch_args.text().strip()
        config.save_option(self.app_name, "pre_launch_command", " ".join([command, arguments]))


LaunchSettingsType = TypeVar("LaunchSettingsType", bound=LaunchSettingsBase)
